from dataclasses import dataclass, field
import os
from typing import Optional, List
from travo.utils import git_get_origin
from travo.gitlab import GitLab, ResourceRef, Project, User, Group

@dataclass
class Homework:
    """
    A homework is a simple aggregator of travo information + configuration.

    The basic logic revolve around

    * The instructor's assignment.
    * Some students' copies that are forks.
    """

    gitlab: GitLab
    project: Project # the original project
    assignment: Project # the assignment (can be `project` or not)
    directory: str # The working directory. Usually `.`

    group: Optional[Group] = None # The possible correction group
    group_level: int = 20 # The access level of the group

    instructor: Optional[User] = None # The possible instructor
    instructor_level: int = 40 # The access level of the instructor


    def __init__(self, url: str = '.') -> None:
        """
        If url is not given, then the remote of the local git repository is used.
        """

        self.project = self.get_project(url)
        self.gitlab = self.project.gitlab
        self.log = self.gitlab.log

    def get_project(self, url: str) -> Project:
        """
        Connect to the forge and get the project.
        """

        if not url.startswith("https:"):
            self.directory = url
            try:
                url = git_get_origin(url)
            except:
                if url == ".":
                    raise RuntimeError(f"The current directory is not a valid travo/git directory. Specify a valid directory or an project URL.")
                else:
                    raise RuntimeError(f"{url} is not a valid travo/git directory. Specify a valid directory or an project URL.")
        else:
            self.directory = '.'

        ref = ResourceRef(url=url)
        path = ref.path
        gitlab = ref.forge

        gitlab.login()

        user = gitlab.get_current_user()
        assert isinstance(user, User)
        gitlab.log.info(f"user: {user.username} ({user.name}) {user.web_url}")

        project = gitlab.get_project(path)
        gitlab.log.info(f"project: '{project.name_with_namespace}' {project.web_url}")
        return project

    def get_group(self, path: str) -> None:
        """
        Find the correction group
        """

        self.group = self.gitlab.get_group(path)

    def get_copies(self) -> List[Project]:
        """
        Get copies.

        If the instructor's assignment is given, get all the students' copies.
        IF a student copy is given, just return it.
        """

        project = self.project

        if project.forked_from_project is not None:
            # This is likely an assignment
            assignment = project.forked_from_project
            self.assignment = assignment
            project.gitlab.log.info(f"Fork of: '{assignment.name_with_namespace}' {assignment.web_url}")
            forks = [project]
            self.forks = forks
            return forks

        self.assignment = project
        forks = project.get_forks(recursive=False)
        self.copies = forks
        return forks


    def check_student(self, project: 'Project', fixup:bool=False) -> bool:
        """
        Check various configuration on a student copy.

        If `fixup` is true then also try to fix them.

        Return `true` if there is issues (after a potential fixup).
        """

        recheck = False
        result = True

        # Detect instructor's assignment if non is given
        assignment = self.assignment
        if assignment is None:
            assignment = project.forked_from_project

        # Check fork
        if assignment is not None:
            if project.forked_from_project is None:
                print(f"  ❌ not a fork")
                result = False
                if fixup:
                    project.add_origin(assignment)
                    recheck = True
            elif assignment.id != project.forked_from_project.id:
                print(f"  ❌ bad fork of {project.forked_from_project.name_with_namespace}")
                result = False

        # Check visibiliy
        vis = project.visibility
        if vis != "private":
            print(f"  ❌ bad visibility {vis}")
            result = False
            if fixup:
                project.setattributes(visibility="private")
                recheck = True

        # TODO: Check allowed file changes

        # Check correction group (if any)
        if self.group is not None:
            sharedok = False
            for g in project.shared_with_groups:
                if g['group_id'] == self.group.id:
                    if g['group_access_level'] >= self.group_level:
                        sharedok = True
                        continue
            if not sharedok:
                print(f"  ❌ bad group sharing with group {self.group.full_path}")
                result = False
                if fixup:
                    project.share_with(self.group, self.group_level)
                    recheck = True

        instructor = self.instructor
        if instructor is None and assignment is not None:
            owners = assignment.get_owners()
            if len(owners) == 0:
                print(f"  ❌ no assignment owner")
                result = False
            elif len(owners) > 1:
                print(f"  ❌ multiple assignment owners {owners}")
                result = False
            else:
                instructor = owners[0]

        if instructor is not None:
            instructok = False
            for member in project.get_members():
                if member['id'] == instructor.id and member['access_level'] >= self.instructor_level:
                    instructok = True
                    continue
            if not instructok:
                print(f"  ❌ bad membership of instructor {instructor.username}")
                result = False
                if fixup:
                    #FIXME share_with do not works on users yet
                    #project.share_with(instructor, self.instructor_level)
                    recheck = True

        if recheck:
            print(f"     Recheck after fixup!")
            result = self.check_student(project)
        return result


    def print_info(self, project: 'Project', fixup:bool=False) -> None:
        """
        Print most information about a student project
        """

        print(f"Student: {project.name_with_namespace} {project.web_url}`")

        # Detect instructor's assignment if non is given
        assignment = self.assignment
        if assignment is None:
            assignment = project.forked_from_project

        self.check_student(project, fixup=fixup)

        if project.default_branch is None:
            print(f"  ❌ empty repository")
            return

        nocommits = False
        compare = None
        if assignment is not None:
            compare = project.get_compare(assignment)
            if 'message' in compare:
                print(f"  ❌ {compare['message']}")
                compare = None
        if compare is not None:
            if len(compare['commits']) == 0:
                print(f"  ❌ no commits")
                nocommits = True

            compare_rev = assignment.get_compare(project)
            if len(compare_rev['commits']) != 0:
                print(f"     not up to date with upstream; lags {len(compare_rev['commits'])}")

        if nocommits:
            return

        reports = project.get_reports()
        for k in reports:
            v = reports[k]
            status = v['status']
            if 'total_count' in v:
                if v['success_count'] == v['total_count'] and status == "success":
                    mark = "✅"
                elif v['success_count'] == 0:
                    mark = "❌"
                else:
                    mark = "  "
                print(f"  {mark} {k}: {v['status']} {v['success_count']}/{v['total_count']}")
            else:
                if status == "success":
                    mark = "✅"
                elif status == "failed":
                    mark = "❌"
                else:
                    mark = "  "
                print(f"  {mark} {k}: {v['status']}")



##############################################################################
# UI and dashboards
#
# Components:
# - AuthenticationWidget
# - LogWidget
# - Grid of AssignmentDashboards
#
# Design notes
#
# - Any call from the UI to the model (in particular any git/gitlab
#   interaction) goes through LogWidget.callback(action) which is in
#   charge of running the actions in subthreads, logging them, and
#   handling failures.
##############################################################################

import io
import glob
import logging
import os
import subprocess
import requests
from threading import Thread
from typing import Any, Callable, Dict, Optional, List, Tuple

try:
    from ipywidgets import (  # type: ignore
        Button,
        ToggleButton,
        GridBox,
        HBox,
        VBox,
        Dropdown,
        Label,
        Layout,
        Output,
        Password,
        Text,
        Textarea,
        RadioButtons,
    )
    import ipywidgets
except ImportError:
    raise ImportError("Cannot find ipywidgets. "
          "dashboards needs optional dependencies. "
          "Please install travo with 'pip install travo[jupyter]'." )

import numpy as np    # type: ignore
import pandas as pd   # type: ignore

from .assignment import Assignment
from .course import Course, CourseAssignment

from .gitlab import GitLab, AuthenticationError, ResourceNotFoundError, unknown
from .utils import run
from travo.i18n import _
from . import jupyter_course    # only used by type hints

# TODO: should use the current foreground color rather than black
border_layout = {'border': '1px solid black'}


def HTML(*args: Any, **kwargs: Any) -> ipywidgets.HTML:
    """
    Make ipywidgets.HTML apply standard styling to its content

    See https://github.com/jupyter-widgets/ipywidgets/issues/2813
    """
    html = ipywidgets.HTML(*args, **kwargs)
    html.add_class('jp-RenderedHTMLCommon')
    return html


class AuthenticationWidget(VBox):
    """
    A widget for interactive input of credentials

    It's meant to be included in the UI; it will display itself upon
    request, send the credentials through forge.login, and hide itself
    if authentication succeeds.

    """
    on_login_hooks: List[Callable]

    def __init__(self, forge: GitLab):
        forge.on_missing_credentials = self.on_missing_credentials
        self.forge = forge
        title = Label(_("authentication widget title", url=forge.base_url))
        self.usernameUI = Text(description=_("username"), layout={'width': 'fit-content'})
        self.passwordUI = Password(description=_("password"), layout={'width': 'fit-content'})
        self.button = Button(description=_("sign in"),
                             button_style="primary"
                             )
        self.messageUI = Label()
        self.button.on_click(lambda event: self.login())
        # We don't want to call login each time a new character is typed in!
        # Is there a way to only observe the "Return" key press?
        # self.passwordUI.observe(lambda event: self.login(),
        #                         names='value')
        self.on_login_hooks = []
        super(AuthenticationWidget, self).__init__(
            children=(title,
                      self.usernameUI,
                      self.passwordUI,
                      self.button,
                      self.messageUI),
            layout={'border': '1px solid black',
                    'display': 'none'})

    def show_widget(self, message: str = "") -> None:
        self.messageUI.value = message
        self.layout.display = 'block'

    def hide_widget(self) -> None:
        self.layout.display = 'none'

    def login(self) -> None:
        self.messageUI.value = ""
        self.button.disabled = True
        username = self.usernameUI.value
        password = self.passwordUI.value
        try:
            self.forge.login(username=username,
                             password=password,
                             anonymous_ok=True)
            self.hide_widget()
            for hook in self.on_login_hooks:
                Thread(target=hook).start()
        except AuthenticationError as e:
            self.messageUI.value = str(e)
        finally:
            self.button.disabled = False

    def on_login(self, hook: Callable) -> None:
        self.on_login_hooks.append(hook)

    def on_missing_credentials(self,
                               forge: GitLab,
                               username: Optional[str] = None,
                               password: Optional[str] = None
                               ) -> Tuple[str, str]:
        """
        Display the widget and raise to interrupt login
        """
        message = _("authentication required")
        self.show_widget(message)
        raise AuthenticationError(message)


class StatusBar(VBox):
    def __init__(self,
                 authentication_widget: AuthenticationWidget,
                 log: logging.Logger
                 ) -> None:
        minimize_layout = {'flex': '0 0 content'}
        self.authentication_widget = authentication_widget
        self.statusUI = Label(_('ready'),
                              layout={'flex': '1 0 content'})
        self.log_show_UI = ToggleButton(description=_('show log details'),
                                        tooltip=_('show command log'),
                                        value=False,
                                        icon="eye",
                                        layout=minimize_layout)
        self.log_level_UI = Dropdown(options=['WARNING', 'INFO', 'DEBUG'],
                                     tooltip=_('command log level'),
                                     layout=minimize_layout)
        self.logUI = Output(layout={'width': 'fit-content'})
        self.logUI.layout.display = 'none'

        def set_log_show(event: Any) -> None:
            if self.log_show_UI.value:
                self.logUI.layout.display = 'block'
                self.log_show_UI.icon = 'eye-slash'
                self.log_show_UI.description = _('hide log details')
            else:
                self.logUI.layout.display = 'none'
                self.log_show_UI.icon = 'eye'
                self.log_show_UI.description = _('show log details')

        self.log_show_UI.observe(set_log_show, names='value')

        def set_log_level(event: Any) -> None:
            log.setLevel(self.log_level_UI.value)

        self.log_level_UI.observe(set_log_level, names='value')

        super().__init__(
            [HBox([Label(_('status') + ': ',
                         layout=minimize_layout),
                   self.statusUI,
                   self.log_show_UI,
                   self.log_level_UI]),
             self.logUI],
            layout=border_layout)

    def run(self,
            action: str,
            command: Callable,
            kwargs: dict = {},
            on_finished: Optional[Callable] = None,
            on_success: Optional[Callable] = None
            ) -> None:
        self.statusUI.value = action + ": " + _("ongoing")
        exit_status = _("failed")  # in case of unexpected exception
        success = False
        with self.logUI:
            try:
                command(**kwargs)
                success = True
                exit_status = _("finished")
            except (RuntimeError, subprocess.CalledProcessError) as e:
                exit_status = _("failed with error", error=str(e))
            finally:
                if on_finished is not None:
                    on_finished()
                self.statusUI.value = action + ": " + exit_status
            if success and on_success is not None:
                on_success()

    def run_in_subthread(self,
                         action: str,
                         command: Callable,
                         on_finished: Optional[Callable] = None,
                         on_success: Optional[Callable] = None
                         ) -> None:
        Thread(target=lambda: self.run(action=action,
                                       command=command,
                                       on_finished=on_finished,
                                       on_success=on_success)).start()


class AssignmentStudentDashboard(HBox):
    def __init__(self,
                 assignment: Assignment,
                 status_bar: Optional[StatusBar] = None,
                 authentication_widget: Optional[AuthenticationWidget] = None
                 ):
        self.name = assignment.name
        self.assignment = assignment
        if authentication_widget is None:
            authentication_widget = AuthenticationWidget(assignment.forge)
        authentication_widget.on_login(self.update)
        self.authentication_widget = authentication_widget
        if status_bar is None:
            status_bar = StatusBar(authentication_widget=authentication_widget,
                                   log=assignment.log)
        self.status_bar = status_bar

        from ipylab import JupyterFrontEnd  # type: ignore
        self.jupyter_front_end = JupyterFrontEnd()

        layout = Layout(width="initial")
        self.nameUI = HTML(self.name, layout=layout)
        self.fetchUI = Button(
            description=_("fetch"),
            button_style="primary",
            icon="download",
            tooltip=_("fetch assignment", assignment_name=self.name),
            layout=layout,
            disabled=True,
        )
        self.fetchUI.on_click(lambda event: self.fetch())
        self.submitUI = Button(
            description=_("submit"),
            button_style="primary",
            icon="upload",
            tooltip=_("submit assignment", assignment_name=self.name),
            layout=layout,
            disabled=True,
        )
        self.submitUI.on_click(lambda event: self.submit())
        self.work_dir_UI = Button(
            description=_("open"),
            button_style="primary",
            icon="edit",
            tooltip=_("open assignment", assignment_name=self.name),
            layout=layout,
            disabled=True,
        )
        self.work_dir_UI.on_click(self.open_work_dir)
        self.scoreUI = HTML(
            "",
            tooltip=_("browse feedback")
        )

        self.submissionUI = HTML(
            layout=Layout(align_self="center"))

        HBox.__init__(
            self,
            [
                self.nameUI,
                self.fetchUI,
                self.work_dir_UI,
                self.submitUI,
                self.submissionUI,
                self.scoreUI,
            ],
        )
        Thread(target=self.update).start()

    def update(self) -> None:
        # For now, fetching the assignment status requires the user to
        # be logged in. Fails gracefuly if this is not yet the case.
        try:
            status = self.assignment.status()
        except AuthenticationError:
            return
        if status.status == "not released":
            self.nameUI.value = self.name
            self.fetchUI.disabled = True
        else:
            repo = self.assignment.repo()
            self.nameUI.value = (
                f'<a href="{repo.web_url}" target="_blank">{self.name}</a>'
            )
            self.fetchUI.disabled = False

        if status.is_submitted():
            personal_repo = self.assignment.personal_repo()
            assert status.team is not None

            def annotation(username: str) -> str:
                result = ''
                assert status.team is not None
                if len(status.team) == 1:
                    return result
                if username == status.student:
                    result += f' ({_("me")})'
                if username == status.leader_name:
                    result += ' *'
                return result

            self.submissionUI.value = \
                '<div style="line-height: 1.1;">' + \
                '<br>'.join(
                    f'<a href="{status.team[username].web_url}"'
                    f'  target="_blank">{username}</a>'
                    f'{annotation(username)}'
                    for username in sorted(status.team.keys())
                ) + \
                '</div>'
        else:
            self.submissionUI.value = ""

        self.submitUI.tooltip = _("submit assignment", assignment_name=self.name)
        self.submitUI.disabled = True
        self.work_dir_UI.disabled = True
        if self.assignment.assignment_dir is not None and os.path.exists(
                self.assignment.assignment_dir
        ):
            self.work_dir_UI.disabled = False
            if self.assignment.submissions_forked_from_path() is unknown \
               and not status.is_submitted():
                # To be generalized: the current tooltip makes the
                # assumption of a course assignments with student
                # groups
                self.submitUI.tooltip += " (" + _("needs student group") + ")"
            else:
                self.submitUI.disabled = False

        if status.autograde_status == "success":
            # these two s/could be provided by autograde_status
            try:
                badge = personal_repo.fetch_artifact(
                    status.autograde_job,
                    artifact_path="feedback/scores.svg").text
            except requests.HTTPError:
                badge = _("browse")
            assert status.autograde_job is not None
            path_components = self.assignment.personal_repo_path().split('/')
            feedback_url = (
                f"https://{path_components[0]}.{self.assignment.forge.base_url[8:]}-/{'/'.join(path_components[1:])}/-/jobs/{status.autograde_job['id']}/artifacts/feedback/scores.html"
            )
            self.scoreUI.value = f"<a href='{feedback_url}' target='_blank'>{badge}</a>"
        else:
            self.scoreUI.value = ""

        # repo = HTML('<a href="#" class="btn btn-primary btn-lg disabled" role="button" aria-disabled="true">Primary link</a>')
        # need to update repo_path

    def fetch(self) -> None:
        self.fetchUI.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("fetching assignment", assignment_name=self.name),
            command=self.assignment.fetch,
            on_finished=lambda: setattr(self.fetchUI, 'disabled', False),
            on_success=self.update
        )

    def submit(self) -> None:
        self.submitUI.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("submitting assignment", assignment_name=self.name),
            command=self.assignment.submit,
            on_finished=lambda: setattr(self.submitUI, 'disabled', False),
            on_success=self.update
        )

    def open_work_dir(self, event: Any) -> None:
        if self.assignment.assignment_dir is None:
            raise ValueError(_("Can't open work dir: assignment_dir is not set"))
        path = os.path.dirname(self.jupyter_front_end.sessions.current_session['path'])
        # TODO : if README.md does not exist neither, try to open gitlab file browser
        index_files = ["index.md", "index.ipynb", "README.md", "README.ipynb"]
        for x in index_files:
            file = os.path.join(self.assignment.assignment_dir, x)
            if os.path.isfile(file):
                self.jupyter_front_end.commands.execute(
                    "docmanager:open",
                    {
                        "path": path + "/" + file,  # Generalize if there is no index.md
                        "factory": "Notebook",
                        # 'options': {
                        #     'mode': 'split-right'
                        # },
                        "kernelPreference": {
                            "shutdownOnClose": True,
                        },
                    },
                )
                return
        raise FileNotFoundError(_("no index file",
                                  assignment_dir=self.assignment.assignment_dir,
                                  index_files=" ".join(index_files)))


class CourseStudentDashboard(VBox):
    """
    A Jupyter-widget based course dashboard for students

    This class currently assumes that the user is logged in.
    """
    student_group_UI: Dropdown
    assignments: Tuple[str, ...] = ()
    assignment_dashboards: Dict[str, AssignmentStudentDashboard]

    def __init__(self,
                 course: Course,
                 student_group: Optional[str] = None):
        self.course = course

        self.header = HBox(
            [HTML(f'<a href="{course.url}" target="_blank">{course.name}</a>',
                  layout={'flex': '1 0 content'})],
            layout=border_layout)
        if self.course.student_groups is not None:
            self.student_group_UI = Dropdown(
                description=_('student group'),
                value=student_group,
                options=course.student_groups,
            )
            self.header.children += (self.student_group_UI,)
            self.student_group_UI.observe(lambda change: self.update_student_group(),
                                          names='value')
        else:
            self.student_group_UI = Dropdown()

        self.authentication_widget = AuthenticationWidget(
            forge=self.course.forge)

        self.grid = GridBox(
            layout=Layout(
                grid_template_columns='repeat(6, max-content)',
                grid_gap='5px 5px',
                border=border_layout['border']
            )
        )
        self.assignment_dashboards = {}

        self.status_bar = StatusBar(authentication_widget=self.authentication_widget,
                                    log=self.course.log)

        super().__init__([self.authentication_widget,
                          self.header,
                          self.grid,
                          self.status_bar],
                         layout={'width': 'fit-content'},
                         )
        self.update(update_assignment_list=True)
        try:
            self.course.forge.login()
        except AuthenticationError:
            pass

    def update_student_group(self) -> None:
        student_group = self.student_group_UI.value
        for assignment_name in self.assignments:
            assignment_dashboard = self.assignment_dashboards[assignment_name]
            assignment = assignment_dashboard.assignment
            assert isinstance(assignment, CourseAssignment)
            assignment.student_group = student_group
            assignment_dashboard.update()

    def update(self, update_assignment_list: bool = False) -> None:
        student_group = self.student_group_UI.value
        # if student_group is None:
        #    self.center = None
        if update_assignment_list:
            if self.course.assignments is not None:
                assignments = self.course.assignments
            else:
                assignments = self.course.get_released_assignments()
            if tuple(assignments) != self.assignments:
                self.assignments = tuple(assignments)
            else:
                update_assignment_list = False

        for assignment in self.assignments:
            if assignment not in self.assignment_dashboards:
                self.assignment_dashboards[assignment] = AssignmentStudentDashboard(
                    self.course.assignment(
                        assignment, student_group=student_group
                    ),
                    status_bar=self.status_bar,
                    authentication_widget=self.authentication_widget
                )

        if update_assignment_list:
            self.grid.children = [
                Label(label)
                for j, label in enumerate(
                        [_('assignment'),
                         '',
                         _('work directory'),
                         '',
                         _('submission'),
                         _('score'),
                         ])
            ] + [widget
                 for assignment in self.assignments
                 for widget in self.assignment_dashboards[assignment].children]


class AssignmentInstructorDashboard(HBox):
    def __init__(self,
                 course: 'jupyter_course.JupyterCourse',
                 assignment: Assignment,
                 force_autograding: bool,
                 new_score_policy: str,
                 student_group: Optional[Any] = None,
                 release_mode: Optional[Any] = None,
                 status_bar: Optional[StatusBar] = None,
                 authentication_widget: Optional[AuthenticationWidget] = None
                 ):
        self.course = course
        self.student_group = student_group
        self.release_mode = release_mode
        self.force_autograding = force_autograding
        self.new_score_policy = new_score_policy
        # TODO: find better way to recover the assignment name
        self.name = assignment.name
        self.assignment = assignment
        if authentication_widget is None:
            authentication_widget = AuthenticationWidget(assignment.forge)
        self.authentication_widget = authentication_widget
        self.authentication_widget.on_login(self.update)
        if status_bar is None:
            status_bar = StatusBar(authentication_widget=authentication_widget,
                                   log=assignment.log)
        self.status_bar = status_bar

        from ipylab import JupyterFrontEnd  # type: ignore
        self.jupyter_front_end = JupyterFrontEnd()

        layout = Layout(width="initial")
        self.nameUI = Label(self.name)
        self.assignmentUI = HTML()  # layout=layout)
        self.generateButton = Button(
            description="generate",
            button_style="primary",
            icon="cog",
            tooltip=_("generate_assignment assignment", assignment_name=self.name),
            layout=layout,
            disabled=False)
        self.generateButton.on_click(lambda event: self.generate())
        self.generateStatus = Textarea(self.generate_status_cmd(), layout=layout)
        self.generateUI = VBox([self.generateButton, self.generateStatus])

        self.releaseButton = Button(
            description="release",
            button_style="primary",
            icon="angle-double-up",
            tooltip=_("release assignment", assignment_name=self.name),
            layout=layout,
            disabled=False,
        )
        self.releaseButton.on_click(lambda event: self.release(self.release_mode))
        self.releaseStatus = Textarea(self.release_status_cmd(), layout=layout)
        self.releaseUI = VBox([self.releaseButton, self.releaseStatus])

        self.collectButton = Button(
            description="collect",
            button_style="primary",
            icon="angle-double-down",
            tooltip=_("collect assignment", assignment_name=self.name),
            layout=layout,
            disabled=False,
        )
        self.collectButton.on_click(lambda event: self.collect())
        self.collectStatus = Textarea(self.count_submissions_cmd(submitted_directory="submitted"), layout=layout)
        self.collectUI = VBox([self.collectButton, self.collectStatus])

        self.formgraderButton = Button(
            description="formgrader",
            button_style="primary",
            icon="book",
            tooltip=_("formgrader assignment", assignment_name=self.name),
            layout=layout,
            disabled=False,
        )
        self.formgraderButton.on_click(lambda event: self.open_formgrader_cmd(event))
        self.formgraderStatus = Textarea(self.count_submissions_need_manual_grade_cmd(),
                                         layout=layout)
        self.formgraderUI = VBox([self.formgraderButton, self.formgraderStatus])

        self.feedbackButton = Button(
            description="feedback",
            button_style="primary",
            icon="paper-plane",
            tooltip=_("feedback assignment", assignment_name=self.name),
            layout=layout,
            disabled=False,
        )
        self.feedbackButton.on_click(lambda event: self.feedback())
        self.feedbackStatus = Textarea(self.release_feedback_status_cmd(self.student_group), layout=layout)
        self.feedbackUI = VBox([self.feedbackButton, self.feedbackStatus])

        HBox.__init__(
            self,
            [
                self.nameUI,
                self.assignmentUI,
                self.generateUI,
                self.releaseUI,
                self.collectUI,
                self.formgraderUI,
                self.feedbackUI,
            ],
        )
        Thread(target=self.update).start()

    def update(self) -> None:
        self.generateStatus.value = self.generate_status_cmd()
        if self.release_mode is None:
            self.releaseButton.tooltip += "(needs release mode)"
            self.releaseButton.disabled = True
        # For now, fetching the assignment status requires the user to
        # be logged in. Fails gracefuly if this is not yet the case.
        try:
            status = self.assignment.status()
        except AuthenticationError:
            return
        if status.status == "not released":
            self.assignmentUI.value = ""
            self.collectButton.disabled = True
            self.feedbackButton.disabled = True
            self.formgraderButton.disabled = True
        else:
            repo = self.assignment.repo()
            self.assignmentUI.value = (
                f'<a href="{repo.web_url}" target="_blank">{_("browse")}</a>'
            )
            # self.generateButton.disabled = False
            self.releaseStatus.value = self.release_status_cmd()
        if self.student_group is None:
            self.collectButton.disabled = True
            self.feedbackButton.disabled = True
        else:
            self.collectButton.disabled = False
            self.feedbackButton.disabled = False
            self.formgraderButton.disabled = False

        self.collectStatus.value = self.count_submissions_cmd(submitted_directory="submitted")
        self.formgraderStatus.value = self.count_submissions_need_manual_grade_cmd()
        self.feedbackStatus.value = self.release_feedback_status_cmd(self.student_group)

    def generate(self) -> None:
        self.generateButton.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("generating assignment", assignment_name=self.name),
            command=self.course.generate_assignment,
            kwargs={"assignment_name": self.name},
            on_finished=lambda: setattr(self.generateButton, 'disabled', False),
            on_success=lambda: self.update
        )

    def release(self, release_mode: Optional[Any] = None) -> None:
        if release_mode not in ["public", "private"]:
            self.course.log.error("Please set release mode: public or private.")
            return
        self.releaseButton.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("releasing assignment", assignment_name=self.name),
            command=self.course.release,
            kwargs={"assignment_name": self.name, "visibility": release_mode},
            on_finished=lambda: setattr(self.releaseButton, 'disabled', False),
            on_success=lambda: self.update
        )

    def collect(self) -> None:
        self.collectButton.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("collect assignment", assignment_name=self.name),
            command=self.collect_cmd,
            on_finished=lambda: setattr(self.collectButton, 'disabled', False),
            on_success=lambda: self.update
        )

    def formgrader(self) -> None:
        self.status_bar.run(  # run_in_subthread(
            action=_("open formgrader", assignment_name=self.name),
            command=self.open_formgrader_cmd,
            kwargs={},
            on_finished=lambda: setattr(self.formgraderButton, 'disabled', False),
            on_success=lambda: self.update
        )

    def feedback(self) -> None:
        self.feedbackButton.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("feedbacking assignment", assignment_name=self.name),
            command=self.feedback_cmd,
            kwargs={"tag": '*'},
            on_finished=lambda: setattr(self.feedbackButton, 'disabled', False),
            on_success=lambda: self.update
        )

    def collect_cmd(self) -> None:
        if self.student_group is None:
            self.course.log.error("Must choose a student_group value.")
            return
        self.course.collect_for_nbgrader(assignment_name=self.name, student_group=self.student_group)
        self.course.ensure_autograded(assignment_name=self.name, student_group=self.student_group,
                                      force_autograde=self.force_autograding)
        self.course.collect_autograded(assignment_name=self.name, student_group=self.student_group)
        self.course.merge_autograded_db(assignment_name=os.path.basename(self.name), on_inconsistency="WARNING", back=False,
                                        new_score_policy=self.new_score_policy)
        self.course.clear_needs_manual_grade(os.path.basename(self.name))

    def feedback_cmd(self, tag: str = '*') -> None:
        if self.student_group is None:
            self.course.log.error("Must choose a student_group value.")
            return
        self.course.generate_feedback(os.path.basename(self.name), tag=tag,
                                      new_score_policy=self.new_score_policy)
        # self.course.merge_autograded_db(os.path.basename(self.name), back=True, on_inconsistency="WARNING",
        #                                tag=tag, new_score_policy=self.new_score_policy.value)
        self.course.release_feedback(self.name, student_group=self.student_group)

    def open_formgrader_cmd(self, event: Any) -> None:
        try:  # jupyterlab
            from ipylab import JupyterFrontEnd  # type: ignore

            # if self.assignment.assignment_dir is None:
            #    raise ValueError("Can't open work dir if assignment_dir is not set")
            path = os.path.dirname(self.jupyter_front_end.sessions.current_session['path'])
            file = self.course.formgrader(os.path.basename(self.name), in_notebook=False)
            import pdb
            pdb.pm()
            self.jupyter_front_end.commands.execute(
                "docmanager:open",
                {
                    "path": os.path.join(path, file),  # Generalize if there is no index.md
                    "factory": "Notebook",
                    # 'options': {
                    #     'mode': 'split-right'
                    # },
                    "kernelPreference": {
                        "shutdownOnClose": True,
                    },
                },
            )
        except KeyError:  # jupyter notebook
            from IPython.display import display   # type: ignore
            display(self.course.formgrader(os.path.basename(self.name), in_notebook=True))   # type: ignore

    def generate_status_cmd(self) -> str:
        try:
            path = os.path.join(self.course.release_directory, os.path.basename(self.assignment.name))
            date = subprocess.run(['git', 'log', '-1', '--format=%ai'], check=True, text=True, capture_output=True,
                                  cwd=path).stdout
            status = f"last generation: {date}"
        except:
            status = "not generated"
        return status

    def count_submissions_cmd(self, submitted_directory: str = "submitted") -> str:
        status = ""
        number_of_copies = len(glob.glob(f"{submitted_directory}/*/{os.path.basename(self.assignment.name)}",
                                         recursive=True))
        if number_of_copies > 0:
            status += f"#submissions: {number_of_copies}\n"
        return status

    def count_submissions_need_manual_grade_cmd(self) -> str:
        status = ""
        try:
            # Create the connection to the database
            from nbgrader.api import Gradebook, MissingEntry  # type: ignore
            with Gradebook('sqlite:///.gradebook.db') as gb:
                grades = []
                counts = 0
                total = 0
                # Loop over each assignment in the database
                for student in gb.students:
                    # Try to find the submission in the database. If it doesn't exist, the
                    # `MissingEntry` exception will be raised, which means the student
                    # didn't submit anything, so we assign them a score of zero.
                    try:
                        submission = gb.find_submission(os.path.basename(self.assignment.name), student.id)
                    except MissingEntry:
                        pass
                    else:
                        total += 1
                        score = submission.score
                        grades.append(score)
                        if submission.needs_manual_grade:
                            counts += 1
                if len(grades) > 0:
                    status += f"Needs manual grading: {counts}\nGrades: {np.mean(grades):.2f}+/-{np.std(grades):.2f}\n"
                else:
                    status = "No grades"
        except:
            pass
        return status

    def release_status_cmd(self) -> str:
        try:
            p = self.course.forge.get_project(self.assignment.repo_path)
            status = f"last release: {p.visibility}\n{p.last_activity_at}"
        except ResourceNotFoundError:
            status = "not released"
        return status

    def release_feedback_status_cmd(self, student_group: Optional[Any]) -> str:
        try:
            url = self.course.forge.get_project(self.course.assignment(self.assignment.name,
                                                                       student_group=student_group).repo_path).http_url_to_repo
            url = "go to " + url.replace(".git", "/-/forks")
        except ResourceNotFoundError:
            url = "not released"
        return url


class CourseInstructorDashboard(VBox):
    """
    A Jupyter-widget based course dashboard for instructors

    This class currently assumes that the user is logged in.
    """

    def __init__(self,
                 course: 'jupyter_course.JupyterCourse',
                 student_group: Optional[str] = None):
        self.course = course
        self.course.forge.login()

        self.release_mode = RadioButtons(description="", value=None, options=["public", "private"],
                                         layout={'width': 'fit-content'})
        self.force_autograding = RadioButtons(description="", value=False, options=[False, True],
                                              layout={'width': 'fit-content'})
        self.new_score_policy = RadioButtons(description="", value="only_greater",
                                             options=["only_empty", "only_greater", "force_new_score"],
                                             layout={'width': 'fit-content'})

        self.header = HBox(
            [HTML(f'<a href="{course.url}" target="_blank">{course.name}</a>',
                  layout={'flex': 'flex-start'}),
             VBox([Label("release mode"), self.release_mode], width="10px"),
             VBox([Label("collect:\nforce new autograde"), self.force_autograding]),
             VBox([Label("new score policy"), self.new_score_policy])],
            layout=Layout(border='1px solid black', grid_gap='5px 40px'))
        if self.course.student_groups is not None:
            self.student_group_UI = Dropdown(
                description="",
                value=student_group,
                options=course.student_groups,
            )
            self.header.children += (VBox([Label(_('student group')), self.student_group_UI]),)
            self.student_group_UI.observe(lambda change: self.update(),
                                          names='value')
        self.release_mode.observe(lambda change: self.update(),
                                  names='value')
        self.new_score_policy.observe(lambda change: self.update(),
                                      names='value')

        self.authentication_widget = AuthenticationWidget(
            forge=self.course.forge)

        self.grid = GridBox(
            layout=Layout(
                grid_template_columns='repeat(7, max-content)',
                grid_gap='5px 5px',
                border=border_layout['border']
            )
        )

        self.status_bar = StatusBar(authentication_widget=self.authentication_widget,
                                    log=self.course.log)

        super().__init__([self.authentication_widget,
                          self.header,
                          self.grid,
                          self.status_bar],
                         layout={'width': 'fit-content'}
                         )
        self.make_grid()
        try:
            self.course.forge.login()
        except AuthenticationError:
            pass

    def update(self) -> None:
        self.make_grid()
        # Once it will be possible to update on the fly the student group of a course
        # assignment (which means updating the assignment repo and the fork_of_fork field,
        # it will be posible to simply update the assignment dashboards:
        # for assignment_dashboard in self.assignment_dashboards:
        #     Thread(target=assignment_dashboard.update).start()

    def make_grid(self) -> None:
        if self.student_group_UI.value is None:
            self.center = None
        if self.course.assignments is None:
            self.course.assignments = self.course.get_released_assignments()
        self.assignment_dashboards = [
            AssignmentInstructorDashboard(
                self.course,
                self.course.assignment(
                    assignment_name, student_group=self.student_group_UI.value
                ),
                student_group=self.student_group_UI.value,
                release_mode=self.release_mode.value,
                force_autograding=self.force_autograding.value,
                new_score_policy=self.new_score_policy.value,
                status_bar=self.status_bar,
                authentication_widget=self.authentication_widget
            )
            for assignment_name in self.course.assignments
        ]
        self.grid.children = [
                                 Label(label)
                                 for j, label in enumerate(
                                    [_('assignment'),
                                     _('assignment repository'),
                                     'Gestion du cours', '',
                                     'Evaluations', '', '',
                                     ])
                             ] + [widget
                                  for assignment_dashboard in self.assignment_dashboards
                                  for widget in assignment_dashboard.children]


class CourseGradeDashboard(VBox):
    """
    A Jupyter-widget based course dashboard for grade managing.

    This class currently assumes that the user is logged in.
    """

    def __init__(self, course: 'jupyter_course.JupyterCourse') -> None:
        self.course = course
        self.course.forge.login()

        layout = Layout(width="initial")
        self.dashboard_grade_filename = "dashboard-grades.csv"
        if self.course.assignments is None:
            self.course.assignments = self.course.get_released_assignments()
        self.assignments_UI = Dropdown(description="assignments", value=None,
                                       options=["all"] + [a for a in self.course.assignments], width="350px")

        self.get_scores_UI = Button(description="update scores",
                                   button_style="primary",
                                   icon="book",
                                   layout=layout)
        self.get_scores_UI.on_click(lambda event: self.get_scores())

        self.clear_csv_UI = Button(description="clear scores",
                                   button_style="primary",
                                   icon="trash",
                                   layout=layout)
        self.clear_csv_UI.on_click(lambda event: self.clear_csv())

        self.copy_UI = Button(description="copy",
                              button_style="primary",
                              icon="clipboard",
                              layout=layout)
        self.copy_UI.on_click(lambda event: self.copy_dataframe())

        self.header = HBox(
            [self.assignments_UI, self.get_scores_UI, self.clear_csv_UI, self.copy_UI],
            layout=border_layout)
        if self.course.student_groups is not None:
            self.student_group_UI = Dropdown(
                description=_('student group'),
                value=None,
                options=["all"] + list(self.course.student_groups),
            )
            self.header.children = (self.student_group_UI,) + self.header.children

        self.authentication_widget = AuthenticationWidget(
            forge=self.course.forge)
        self.make_grid()
        self.status_bar = StatusBar(authentication_widget=self.authentication_widget,
                                    log=self.course.log)

        super().__init__([self.authentication_widget,
                          self.header,
                          self.grid,
                          self.status_bar],
                         layout={'width': 'fit-content'},
                         )

        try:
            self.course.forge.login()
        except AuthenticationError:
            pass

    def update(self) -> None:
        self.refresh_scores()
        self.grid.data = self.df

    def make_grid(self) -> None:
        from ipydatagrid import DataGrid, TextRenderer, Expr  # type: ignore
        self.df = self.refresh_scores()

        renderers = {}
        for col in self.df.columns:
            if "note" in col and "/0" not in col:
                renderers[col] = TextRenderer(text_color="black", background_color=Expr(_format_note))
            elif "status" in col:
                renderers[col] = TextRenderer(text_color="black", background_color=Expr(_format_status))

        grid_layout = {"height": "300px"}
        self.grid = DataGrid(self.df, base_column_size=200, base_header_size=200, layout=grid_layout, editable=True,
                             renderers=renderers)
        self.grid.auto_fit_params = {"area": "all", "padding": 100}
        self.grid.auto_fit_columns = True

    def refresh_scores(self) -> Any:  # pandas.DataFrame
        if os.path.isfile(self.dashboard_grade_filename):
            df = pd.read_csv(self.dashboard_grade_filename, index_col=0)
        else:
            df = pd.DataFrame()
        col_students = ["group", "email"]
        col_assignments = [col for col in df.columns if "-" in col]
        self.df = df.reindex(col_students + sorted(col_assignments), axis=1)
        return self.df

    def get_scores(self) -> None:
        self.get_scores_UI.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("get scores"),
            command=self.get_scores_cmd,
            on_finished=lambda: setattr(self.get_scores_UI, 'disabled', False),
            on_success=lambda: self.update
        )

    def clear_csv(self) -> None:
        self.clear_csv_UI.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("clear grades"),
            command=self.clear_csv_cmd,
            on_finished=lambda: setattr(self.clear_csv_UI, 'disabled', False),
            on_success=lambda: self.update
        )

    def copy_dataframe(self) -> None:
        self.copy_UI.disabled = True
        self.status_bar.run(  # run_in_subthread(
            action=_("clear grades"),
            command=self.copy_cmd,
            on_finished=lambda: setattr(self.copy_UI, 'disabled', False),
            on_success=None
        )

    def get_scores_cmd(self) -> None:
        """Load all the grades for an assignment and student group.

        Grades are always exported in file dashboard-grades.csv.

        """
        # load the requested assignments
        if self.assignments_UI.value is None or self.assignments_UI.value == "all":
            assignments = [assignment_name for assignment_name in self.assignments_UI.options[1:]]
        else:
            assignments = [self.assignments_UI.value]
        # load the requested groupes
        if self.student_group_UI.value is None or self.student_group_UI.value == "all":
            student_groups = [student_group for student_group in self.student_group_UI.options[1:]]
        else:
            student_groups = [self.student_group_UI.value]

        # load previous scores from dashboard-grades.csv
        d = {}
        if os.path.isfile(self.dashboard_grade_filename):
            df2 = pd.read_csv(self.dashboard_grade_filename, index_col=0)
            d = df2.to_dict("index")

        self.course.forge.login()
        for assignment in assignments:
            total = 0
            total_bad = 0
            for student_group in student_groups:
                # path = course.assignment_repo_path(assignment, student_group)
                try:
                    # project = course.forge.get_project(path)
                    # forks = project.get_forks(recursive=True)
                    submissions_status = self.course.assignment(assignment_name=assignment,
                                                                student_group=student_group).collect_status()
                    # course.log.info(f"Collecting autograded for {len(submissions_status)} students")
                    bad_projects = []
                    if len(submissions_status) == 0:
                        continue
                    for status in submissions_status:
                        student = status.student
                        if student is not None:
                            if student not in d.keys():
                                d[student] = {"group": student_group}
                                # for assignment in assignments:
                            d[student]["email"] = student + "@" + self.course.mail_extension
                            d[student][assignment + "-status"] = "as " + student_group + " with autograding=" + status.autograde_status
                            if status.autograde_status != "success":
                                self.course.log.info(f"missing successful autograde for {student}")
                                bad_projects.append(student)
                                continue
                            self.course.log.info(f"fetching scores for {student}")
                            if status.submission is not None:
                                repo = status.submission.repo
                                job = status.autograde_job
                                path = f"feedback/scores.csv"
                                scores_txt = repo.fetch_artifact(job, artifact_path=path).text
                                scores = pd.read_csv(io.StringIO(scores_txt))
                                total_score = np.sum(scores["total_score"].values)
                                manual_score = np.nansum(scores["manual_score"].values)
                                max_manual_score = np.sum(scores["max_manual_score"].values)
                                if np.all(np.isnan(scores["manual_score"].values)) and ~np.isnan(max_manual_score) and max_manual_score > 0:
                                    d[student][assignment + f"-note (/{int(scores['max_total_score'].values[0])})"] = "attente notation manuelle"
                                else:
                                    d[student][assignment + f"-note (/{int(scores['max_total_score'].values[0])})"] = total_score
                            else:
                                self.course.log.info(f"missing submission for {student}")
                                bad_projects.append(student)
                                continue
                    if len(bad_projects) > 0:
                        self.course.log.warning(f"{len(bad_projects)} dépôt(s) avec autograde_status!=success:")
                        for student in bad_projects:
                            self.course.log.warning(f"student={student}, status={d[student][assignment + '-status']}")
                    total += len(submissions_status)
                    total_bad += len(bad_projects)
                    self.course.log.warning(
                        f"{len(submissions_status)} projets soumis pour {assignment}, groupe {student_group},"
                        f" {len(bad_projects)} dépôt(s) corrompu(s).")
                except ResourceNotFoundError:
                    pass
            self.course.log.info(f"{total} projets soumis pour {assignment}, {total_bad} dépôt(s) corrompu(s).")

        self.df = pd.DataFrame.from_dict(d, orient="index")
        self.df.index.name = "student_id"
        self.df.to_csv(self.dashboard_grade_filename)
        self.df = self.refresh_scores()
        self.update()

    def clear_csv_cmd(self) -> None:
        run(["rm", "-rf", self.dashboard_grade_filename])
        self.update()

    def copy_cmd(self) -> None:
        df2 = self.df.replace("attente notation manuelle", np.nan)
        df2.to_clipboard(excel=True)


def _format_note(cell: Any) -> str:
    return ("yellow" if cell.value == "attente notation manuelle" else "white")


def _format_status(cell: Any) -> str:
    if "autograding=success" in cell.value:
        return "lightgreen"
    elif "autograding=none" in cell.value:
        return "red"
    else:
        return "white"



import os

from travo.assignment import Assignment


def test_fetch_from_empty_personal_repo(standalone_assignment: Assignment,
                                        standalone_assignment_dir: str) -> None:
    assignment = standalone_assignment
    assignment_dir = standalone_assignment_dir
    forge = assignment.forge
    repo = forge.get_project(assignment.repo_path)

    # "Accidently" create an empty personal repository with no fork relation
    my_repo = forge.ensure_project(path=assignment.personal_repo_path(),
                                   name=assignment.personal_repo_name())
    assert my_repo.forked_from_project is None

    # Fetch + submit should recover smoothly
    assignment.fetch(assignment_dir)

    # Content should be recovered from the original repository
    assert os.path.isfile(os.path.join(assignment_dir,
                                       "README.md"))

    assignment.submit(assignment_dir)

    # The personal repo should now have a single branch named master,
    # and be a fork of the assignment repository
    my_repo = forge.get_project(path=assignment.personal_repo_path())
    branch, = my_repo.get_branches()
    assert branch["name"] == "master"
    assert my_repo.forked_from_project is not None
    assert my_repo.forked_from_project.id == repo.id

    # Tear down
    assignment.remove_personal_repo(force=True)



import logging
import os.path
from dataclasses import dataclass, field
from deprecation import deprecated  # type: ignore
import re
import subprocess
from typing import List, Optional, Tuple, Union

from .utils import getLogger, run
from .gitlab import Forge, GitLabTest, Project, Resource, ResourceNotFoundError,\
    unknown, Unknown, AnonymousUser
from .assignment import Assignment
from .i18n import _
from . import __version__


def missing_course() -> 'Course':
    raise ValueError("missing required argument: 'course'")


"""
Characters that are forbidden in GitLab group names

https://docs.gitlab.com/ee/user/reserved_names.html#limitations-on-project-and-group-names

Test:

    >>> re.sub(forbidden_in_gitlab_group_name, " ", "a-s+d o_98#(*&$'sadf.)")
    '-a+sd o_98 (    sadf.)'
"""
forbidden_in_gitlab_group_name = re.compile(r"[^.()\w+-]")


@dataclass
class CourseAssignment(Assignment):
    # Until Python 3.10 and keyword only fields, a subdataclass
    # can't add mandatory arguments. We fake `course` being
    # mandatory by providing a default factory raising an error.
    # https://medium.com/@aniscampos/python-dataclass-inheritance-finally-686eaf60fbb5

    course: 'Course' = field(default_factory=missing_course)
    student_group: Optional[str] = None                       # Meant to be mutable

    def personal_repo_path_components(self,
                                      username: Optional[str] = None
                                      ) -> Tuple[str, ...]:
        """
        Return the components from which the path of the student's submission is built

        Example:

            >>> course = getfixture("rich_course")
            >>> assignment = course.assignment("SubCourse/Assignment1")

            >>> assignment.personal_repo_path_components()
            ('travo-test-etu', 'TestCourse', '2020-2021', 'SubCourse', 'Assignment1')

            >>> assignment.personal_repo_path_components(username="john.doo")
            ('john.doo', 'TestCourse', '2020-2021', 'SubCourse', 'Assignment1')

            >>> course.group_submissions = True
            >>> assignment.personal_repo_path_components(username="john.doo")
            ('john_doo_travo', 'TestCourse', '2020-2021', 'SubCourse', 'Assignment1')

            >>> course.path='TestModule/TestCourse'
            >>> assignment.personal_repo_path_components()
            ('travo-test-etu_travo', 'TestModule', 'TestCourse', '2020-2021', 'SubCourse', 'Assignment1')
        """
        root = self.get_username(username)
        if self.course.group_submissions:
            root = root.replace(".", "_") + "_travo"
        components = [root, *self.course.path.split('/')]
        if self.course.session_path is not None:
            components.append(self.course.session_path)
        components.extend(self.name.split('/'))
        return tuple(components)

    def personal_repo_path(self,
                           username: Optional[str] = None
                           ) -> str:
        """
        Return the path on the forge of the student's submission for this assignment

        Examples:

            >>> course = getfixture("course")
            >>> course.assignment("SubCourse/Assignment1").personal_repo_path()
            'travo-test-etu/TestCourse-SubCourse-Assignment1'

            >>> course = getfixture("rich_course")
            >>> course.assignment("SubCourse/Assignment1").personal_repo_path()
            'travo-test-etu/TestCourse-2020-2021-SubCourse-Assignment1'
            >>> course.assignment("SubCourse/Assignment1",
            ...                   student_group="Group1").personal_repo_path()
            'travo-test-etu/TestCourse-2020-2021-SubCourse-Assignment1'

        More examples with grouped submissions:

            >>> course.group_submissions = True
            >>> assignment = course.assignment("SubCourse/Assignment1")
            >>> assignment.personal_repo_path()
            'travo-test-etu_travo/TestCourse/2020-2021/SubCourse/Assignment1'
            >>> assignment.personal_repo_path(username="john.doo")
            'john_doo_travo/TestCourse/2020-2021/SubCourse/Assignment1'
        """
        components = self.personal_repo_path_components(username)
        if self.course.group_submissions:
            return '/'.join(components)
        else:
            return components[0] + '/' + '-'.join(components[1:])

    def personal_repo_name_components(self) -> Tuple[str, ...]:
        """
        Return the components from which the path of the student's submission is built

        Precondition: the user must be logged in, non anymously.

        Examples:

            >>> course = getfixture("rich_course")
            >>> course.forge.login()
            >>> assignment = course.assignment("SubCourse/Assignment1")

            >>> assignment.personal_repo_name_components()
            ('Étudiant de test pour travo',
             'Test course', '2020-2021', 'SubCourse', 'Assignment1')

            >>> course.group_submissions = True
            >>> assignment.personal_repo_name_components()
            (...Étudiant de test pour travo...,
             'Test course', '2020-2021', 'SubCourse', 'Assignment1')

        Test:

            >>> name = assignment.personal_repo_name_components()[0]
            >>> from travo.i18n import _
            >>> expected = _('submission group name', 
            ...              name='Étudiant de test pour travo')
            >>> assert name == expected
        """
        user = self.forge.get_current_user()
        assert not isinstance(user, AnonymousUser)
        assert user.name is not None
        name = user.name

        # Replace forbidden characters by spaces
        name = re.sub(forbidden_in_gitlab_group_name, " ", name)

        if self.course.group_submissions:
            name = _('submission group name', name=user.name)
        components = [name, self.course.name]
        if self.course.session_name:
            components.append(self.course.session_name)
        components.extend(self.name.split('/'))
        return tuple(components)

    def personal_repo_name(self) -> str:
        """
        Return the name of the student's personal repository for the given assignment

        Precondition: the user must be logged in, non anymously.

        Example:

            >>> course = getfixture("course")
            >>> course.forge.login()
            >>> course.assignment("SubCourse/Assignment1").personal_repo_name()
            'Test course - SubCourse - Assignment1'

            >>> course = getfixture("rich_course")
            >>> course.forge.login()
            >>> course.assignment("SubCourse/Assignment1").personal_repo_name()
            'Test course - 2020-2021 - SubCourse - Assignment1'
        """
        components = self.personal_repo_name_components()
        if self.course.group_submissions:
            return components[-1]
        else:
            return ' - '.join(components[1:])

    def ensure_personal_repo(self,
                             leader_name: Optional[str] = None) -> Project:
        """
        Return the personal repository for this assignment

        Creating it and configuring it if needed.

        """
        if self.course.group_submissions:
            path_components = self.personal_repo_path_components()
            name_components = self.personal_repo_name_components()
            current_path = ""
            for (path, name) in list(zip(path_components, name_components))[:-1]:
                if current_path:
                    current_path += "/"
                current_path += path
                self.forge.ensure_group(current_path,
                                        name=name,
                                        visibility="private")
        return super().ensure_personal_repo(leader_name=leader_name)

    def submissions_forked_from_path(self) -> Union[str, Unknown]:
        """Return the path of the repository that submissions should be a fork of.

        If the course has student groups and the student group is not
        specified, then the student repo should be a fork of some
        unknown fork of `repo`. We won't have enough information to
        create the student submission, but if it already exists, we
        can still fetch, submit, etc.

            >>> course = getfixture("rich_course")
            >>> course.assignment("SubCourse/Assignment1").submissions_forked_from_path()
            unknown
            >>> course.assignment("SubCourse/Assignment1",
            ...                   student_group="Group1").submissions_forked_from_path()
            'TestCourse/2020-2021/SubCourse/Group1/Assignment1'
        """
        if self.leader_name is not None:
            return super().submissions_forked_from_path()
        if self.course.student_groups is not None and self.student_group is None:
            return unknown
        return self.course.assignment_repo_path(self.name,
                                                student_group=self.student_group)

    def submissions_forked_from_missing(self) -> None:
        """Callback when forked_from must be known but is not
        """
        assert self.course.student_groups is not None and self.student_group is None
        self.course.check_student_group(self.student_group)

    def submissions_search_from(self) -> Tuple[Project, int]:
        """Return a project `p` and an int `d` such that the submissions for
        this assignment are all the forks of `p` of depth `d`
        """
        path = self.submissions_forked_from_path()
        if path is unknown:
            return (self.repo(), 2)
        else:
            repo = self.forge.get_project(path)
            return (repo, 1)

    def get_submission_username(self, project: Project) -> Optional[str]:
        """Return the username for the given submission

        Example:

            >>> course = getfixture("course")
            >>> assignment_path = getfixture("assignment_path")
            >>> assignment = course.assignment(assignment_path)
            >>> assignment_personal_repo = getfixture("assignment_personal_repo")

            >>> assignment.get_submission_username(assignment.repo())
            >>> assignment.get_submission_username(assignment_personal_repo)
            'travo-test-etu'

        TODO: test with a rich course and the assignment fork for a student group
        """
        if project.path_with_namespace.startswith(self.course.assignments_group_path):
            return None
        else:
            return project.get_creator().username


@dataclass
class Course:
    """Model a course

    Example:

        >>> from travo.gitlab import GitLab
        >>> GitLab.home_dir = getfixture('tmp_path')  # for CI

        >>> forge = GitLab("https://gitlab.dsi.universite-paris-saclay.fr")
        >>> course = Course(forge=forge,
        ...                 path="Info111",
        ...                 name="Info 111 Programmation Impérative",
        ...                 session_path="2022-2023",
        ...                 student_dir="~/ProgImperative",
        ...                 student_groups=["MI1", "MI2", "MI3"],
        ...                 expires_at="2023-12-31",
        ...                 mail_extension="universite-paris-saclay.fr")

    - `expires_at` (optional): date at which instructors will loose access
      to the student repositories, as a YEAR-MONTH-DAY date string.

    With this example, assignments will be stored in Info111/2022-2023,
    with one fork for each student group in, e.g. Info111/2022-2023/MI3

       >>> course.assignments_group_path
       'Info111/2022-2023'
       >>> course.assignments_group_name
       '2022-2023'
       >>> assignment = course.assignment("Semaine2", student_group="MI3")
       >>> assignment.repo_path
       'Info111/2022-2023/Semaine2'
       >>> assignment.submissions_forked_from_path()
       'Info111/2022-2023/MI3/Semaine2'

    If you wish to use another course layout, you can set the
    above variables directly.
    """
    forge:           Forge
    path:            str
    name:            str
    student_dir:     str
    assignments_group_path: str = ""
    assignments_group_name: str = ""
    session_path:    Optional[str] = None
    session_name:    Optional[str] = None  # Will be defined as soon as session_path is
    assignments:     Optional[List[str]] = None
    student_groups:  Optional[List[str]] = None
    script:          str = "travo"
    url:             Optional[str] = None
    jobs_enabled_for_students: bool = False
    log:             logging.Logger = field(default_factory=getLogger)
    mail_extension:  str = ""
    expires_at:      Optional[str] = None
    group_submissions: bool = False

    def __post_init__(self) -> None:
        # TODO: "Check that: name contains only letters, digits, emojis, '_', '.', dash, space. It must start with letter, digit, emoji or '_'."
        if self.session_path is not None:
            self.assignments_group_path = os.path.join(
                self.path,
                self.session_path)
            if self.session_name is None:
                self.session_name = self.session_path

        if self.session_name is not None:
            self.assignments_group_name = self.session_name

        if not self.assignments_group_path:
            self.assignments_group_path = self.path
        if not self.assignments_group_name:
            self.assignments_group_name = os.path.basename(self.assignments_group_path)

    def work_dir(self,
                 assignment: Optional[str] = None,
                 role: str = "student") -> str:
        """
        Return the absolute work directory for all (git) commands

        Examples:

        Let's create a dummy course::

            >>> forge = getfixture("gitlab")
            >>> course = Course(forge=forge,
            ...                 path="...", name="...",
            ...                 student_dir="~/ProgImperative")

        The work directory for a student, for example to clone a new
        assignment, is given by the `student_dir` attribute of the
        course, with the home dir ("~/") expanded:

            >>> course.work_dir(role="student")
            '/.../ProgImperative'

        (where ... is the student's home directory). To work inside a
        given assignment, for example to run `git push` or `git
        pull`, the work directory is::

            >>> course.work_dir(role="student", assignment="Week1")
            '/.../ProgImperative/Week1'

        When `student_dir` is set to ".", the user is in charge of
        being in the appropriate directory for the current operation.
        So this always return "." for the current directory::

            >>> course = Course(forge=forge,
            ...                 path="...", name="...",
            ...                 student_dir=".")
            >>> course.work_dir(role="student")
            '.'
            >>> course.work_dir(role="student", assignment="Week1")
            '.'

        .. CAVEAT::

            This default implementation follows the convention that
            the work directory for an assignment is obtained by
            joining the assignment name to the root work directory.
            Some methods (e.g. assignment_clone) assume that this
            convention is followed and will need to be generalized
            should some course want to use another one.
        """
        assert role == "student"
        dir = self.student_dir
        if dir == ".":
            return "."
        if dir[:2] == "~/":
            dir = os.path.join(self.forge.home_dir, dir[2:])
        if assignment is not None:
            dir = os.path.join(dir, assignment)
        return dir

    def ensure_work_dir(self) -> str:
        """
        Ensure the existence of the student's work directory

        Return the work directory.

        Examples:

            >>> import os.path
            >>> course = getfixture("course")

            >>> work_dir = course.work_dir(); work_dir
            '/.../TestCourse'
            >>> assert not os.path.exists(work_dir)

            >>> course.ensure_work_dir()
            '/tmp/.../TestCourse'

            >>> assert os.path.isdir(work_dir)

        This is an idempotent operation:

            >>> course.ensure_work_dir()
            '/tmp/.../TestCourse'
            >>> assert os.path.isdir(work_dir)
        """
        work_dir = self.work_dir()
        if not os.path.isdir(work_dir):
            self.log.info(_("creating work dir", work_dir=work_dir))
            assert os.path.isabs(work_dir)
            run(["mkdir", "-p", work_dir])
        return work_dir

    def check_student_group(self,
                            student_group: Optional[str],
                            none_ok: bool = False
                            ) -> None:
        """
        Check that the given student group name is valid

        Raise on error; otherwise return nothing.
        """
        if self.student_groups is None or student_group in self.student_groups:
            return
        if none_ok and student_group is None:
            return
        message = ""
        if student_group is not None:
            message += _('unknown group', student_group=student_group) + "\n"
        message += _('specify group',
                     student_groups=', '.join(self.student_groups)) + "\n"
        # message += _('help', script=self.script)

        raise RuntimeError(message)

    def check_assignment(self, assignment: str) -> None:
        """
        Check whether assignment is a valid assignment

        This current default implementation does nothing.
        Alternatively, it could check whether the assignment exists on
        the forge.

        Courses may override it.
        """
        pass

    def assignment_repo_path(self,
                             assignment: str,
                             student_group: Optional[str] = None) -> str:
        """
        Return the path on the forge for the repository holding the
        student version of the given assignment.

        If `group` is provided, then the path of the student_groups' fork
        thereof is returned instead.

        This method may typically be overriden by the course.

        Example:

            >>> course = getfixture("course")
            >>> course.assignment_repo_path("Assignment1")
            'TestCourse/2020-2021/Assignment1'
            >>> course.assignment_repo_path("Subcourse/Assignment1")
            'TestCourse/2020-2021/Subcourse/Assignment1'

            >>> course.assignment_repo_path("Assignment1", student_group="MI1")
            'TestCourse/2020-2021/MI1/Assignment1'
            >>> course.assignment_repo_path("Subcourse/Assignment1", student_group="MI1")
            'TestCourse/2020-2021/Subcourse/MI1/Assignment1'
        """
        result = [self.assignments_group_path]
        dirname  = os.path.dirname(assignment)
        assignment = os.path.basename(assignment)
        if dirname:
            result.append(dirname)
        if student_group:
            result.append(student_group)
        result.append(assignment)
        return '/'.join(result)

    def assignment_repo_name(self, assignment: str) -> str:
        """
        Return the name of the student repository for the given assignment

        This method may typically be overriden by the course.

        Example:

            >>> course = getfixture("course")
            >>> course.assignment_repo_name("SubCourse/Assignment1")
            'Assignment1'
        """
        return os.path.basename(assignment)

    @deprecated(deprecated_in="0.2", removed_in="1.0",
                current_version=__version__,
                details="Use course.assignment(...).personal_repo_path() instead")
    def assignment_personal_repo_path(self,
                                      assignment_name: str,
                                      username: Optional[str] = None
                                      ) -> str:
        """
        Return the path on the forge of the student's personal repository for the given assignment
        """
        return self.assignment(assignment_name,
                               username=username).personal_repo_path()

    @deprecated(deprecated_in="0.2", removed_in="1.0",
                current_version=__version__,
                details="Use course.assignment(...).personal_repo_name() instead")
    def assignment_personal_repo_name(self, assignment_name: str) -> str:
        """
        Return the name of the student's personal repository for the given assignment
        """
        return self.assignment(assignment_name).personal_repo_name()

    Assignment = CourseAssignment

    def assignment(self,
                   assignment_name: str,
                   student_group: Optional[str] = None,
                   leader_name: Optional[str] = None,
                   username: Optional[str] = None
                   ) -> CourseAssignment:
        """
        Return the assignment `assignment` for the given user

        By default, the user is the current user.
        """
        # The path of the original assignment; not the fork of the student groups
        repo_path = self.assignment_repo_path(assignment_name)
        self.check_student_group(student_group,
                                 none_ok=True)
        return self.Assignment(
                          forge=self.forge,
                          course=self,
                          log=self.log,
                          name=assignment_name,
                          instructors_path=self.assignments_group_path,
                          script=self.script,
                          expires_at=self.expires_at,
                          repo_path=repo_path,
                          student_group=student_group,
                          leader_name=leader_name,
                          username=username,
                          assignment_dir=self.work_dir(assignment_name),
                          jobs_enabled_for_students=self.jobs_enabled_for_students,
                          )

    @deprecated(deprecated_in="0.2", removed_in="1.0",
                current_version=__version__,
                details="Use course.assignment(...).personal_repo() instead")
    def assignment_personal_repo(self,
                                 assignment_name: str,
                                 student_group: Optional[str] = None) -> Project:
        """
        Return the personal repository for this assignment
        """
        assignment = self.assignment(assignment_name,
                                     student_group=student_group)
        return assignment.ensure_personal_repo()

    def remove_assignment_personal_repo(self,
                                        assignment: str,
                                        force: bool = False) -> None:
        """
        Remove the users' personal repository for this assignment
        """
        self.assignment(assignment).remove_personal_repo(force=force)

    def fetch(self, assignment: str,
              student_group: Optional[str] = None,
              force: bool = False) -> None:
        """
        fetch the given assignment
        """
        self.ensure_work_dir()
        assignment_dir = self.work_dir(assignment=assignment)
        return self.assignment(assignment, student_group=student_group).fetch(assignment_dir=assignment_dir,
                                                                              force=force)

    def submit(self,
               assignment: str,
               student_group: Optional[str] = None,
               leader_name: Optional[str] = None) -> None:
        """
        submit the given assignment
        """
        self.check_assignment(assignment)
        the_assignment = self.assignment(assignment,
                                         student_group=student_group,
                                         leader_name=leader_name)
        assignment_dir = self.work_dir(assignment)
        the_assignment.submit(assignment_dir=assignment_dir,
                              leader_name=leader_name)

    def share_with(self,
                   assignment_name: str,
                   username: str,
                   access_level: Union[int, Resource.AccessLevels] = Resource.AccessLevels.DEVELOPER,
                   ) -> None:
        try:
            repo = self.assignment(assignment_name).personal_repo()
        except ResourceNotFoundError:
            raise RuntimeError(_("no submission; please submit", assignment_name=assignment_name))
        user = self.forge.get_user(username)
        repo.share_with(user, access=access_level)

    def release(self, assignment: str, visibility: str = "public") -> None:
        """
        Release the student version of the given assignment on the forge

        Assumption: the current working directory holds the student
        version of the assignment, as a git repository

        This sets up the student repository on the forge, if it does
        not yet exist, and pushes all branches to this repository
        (`push --all`).

        In addition, for each student group, it sets up a fork of the
        above directory and pushes there.

        .. CAVEAT::

            Backward incompatible change with the shell version:
            Release is now meant to be run from the assignment
            directory rather than the root directory of the course, to
            make no hypothesis on the layout of the latter.
        """

        self.check_assignment(assignment)
        self.log.info(f"Publish the assignment {assignment} with visibility {visibility}.")

        # travo_gitlab_remove_project "${repo}"
        attributes = dict(
            visibility = visibility,
            issues_enabled = False,
            merge_requests_enabled = False,
            container_registry_enabled = False,
            wiki_enabled = False,
            snippets_enabled = False,
            lfs_enabled = False,
        )

        path = self.assignment_repo_path(assignment)
        name = self.assignment_repo_name(assignment)
        project = self.forge.ensure_project(path=path, name=name, **attributes)
        try:
            project.unprotect_branch(project.default_branch)
        except RuntimeError:
            pass
        self.log.info(f"- Publishing to {path}.")
        self.forge.git(["push", "--all", project.http_url_to_repo])

        if self.student_groups is None:
            return

        for student_group in self.student_groups:
            path = self.assignment_repo_path(assignment, student_group=student_group)
            name = self.assignment_repo_name(assignment)
            self.log.info(f"- Publishing to the student group {student_group}' fork {path}.")
            self.forge.ensure_group(os.path.dirname(path),
                                    name=student_group,
                                    visibility="public")

            fork = project.ensure_fork(path=path,
                                       name=name,
                                       jobs_enabled=False,
                                       **attributes)
            try:
                fork.unprotect_branch(project.default_branch)
            except RuntimeError:
                pass
            self.forge.git(["push", "--all", fork.http_url_to_repo])

    def remove_assignment(self, assignment: str, force: bool = False) -> None:
        """
        Remove the assignment on the forge (DANGEROUS!)

        This is an irreversible operation!
        """
        self.forge.login()
        self.check_assignment(assignment)
        self.log.info(f"Unpublish the assignment {assignment}.")

        if self.student_groups is not None:
            for student_group in self.student_groups:
                a = self.assignment(assignment, student_group=student_group)
                path = a.submissions_forked_from_path()
                assert path is not unknown
                try:  # Don't fail if this fork does not exist (any more)
                    repo = self.forge.get_project(path)
                except ResourceNotFoundError:
                    pass
                else:
                    self.forge.remove_project(path, force=force)

        repo = self.assignment(assignment).repo()
        # Only the main repo has pipelines set
        repo.remove_pipelines()
        self.forge.remove_project(repo.path_with_namespace, force=force)

    def collect(self,
                assignment:    str,
                student_group: Optional[str] = None,
                student:       Optional[str] = None,
                template:      str = "{username}",
                date:          Optional[str] = None) -> None:
        """
        Collect the student's assignments

        Examples:

        Collect all students submissions in the current directory::

            course.collect("Assignment1")

        Collect all students submissions for a given student group,
        laying them out in nbgrader's format, with the student's group
        appended to the username:

            course.collect("Assignment1",
                           student_group="MP2",
                           template="exchange/Course/submitted/MP2-{username}/{path}")
        """
        path = self.assignment_repo_path(assignment, student_group)
        self.forge.collect_forks(path,
                                 username=student,
                                 template=template,
                                 date=date)

    def ensure_instructor_access(self,
                assignment:    str,
                student_group: Optional[str] = None) -> None:
        """
        Ensure instructor access to the student repositories.

        """
        self.forge.login()
        path = self.assignment_repo_path(assignment, student_group)
        project = self.forge.get_project(path)
        forks = project.get_forks(recursive=True)
        for fork in forks:
            self.log.info(_("ensure instructor access", path=fork.path_with_namespace))
            instructors = self.forge.get_group(self.assignments_group_path)
            member_ids = [user['id'] for user in fork.get_members()]
            for instructor in instructors.get_members():
                if instructor.id not in member_ids:
                    fork.share_with(instructor,
                                    access=fork.AccessLevels.MAINTAINER,
                                    expires_at=None)

    def run(self, *args: str) -> subprocess.CompletedProcess:
        """Run an arbitrary shell command"""
        return run(args)

    def get_released_assignments(self, level: str = "", order_by: str = "created_at") -> List[str]:
        group_path = self.assignments_group_path
        if level:
            group_path += "/" + level
            level += "/"
        group = self.forge.get_group(group_path)
        projects = group.get_projects(simple=True,
                                      with_shared=False,
                                      order_by=order_by,
                                      sort='asc')
        return [level + p.path for p in projects]



import os.path
import pkg_resources
import typing
import i18n      # type: ignore
i18n.set('enable_memoization', True)
i18n.set('fallback', 'en')
i18n.load_path.append(pkg_resources.resource_filename(__name__, '/locale'))
i18n.set('locale', os.environ.get('LANG', 'en')[:2])


def _(key: str, **kwargs: str) -> str:
    """Return the i18n translation for this key

    Examples:

        >>> from travo.i18n import _
        >>> import i18n

        >>> i18n.set('locale', 'en')
        >>> _('hi')
        'Hello'
        >>> _('help', script='foo')
        "Type 'foo' for help"

        >>> i18n.set('locale', 'fr')
        >>> _('hi')
        'Bonjour'
        >>> _('help', script='foo')
        "Tapez «foo» pour de l'aide"
    """
    return typing.cast(str, i18n.t(f'travo.{key}', **kwargs))



import os
import subprocess
import urllib
import urllib.parse
from typing import cast, Any, Sequence, Optional
import logging
import colorlog # type: ignore

_logger : Optional[logging.Logger] = None
def getLogger() -> logging.Logger:
    global _logger
    if _logger is None:
        handler = colorlog.StreamHandler()
        handler.setFormatter(colorlog.ColoredFormatter(
            '%(log_color)s%(levelname)s:%(name)s:%(message)s'))
        _logger = cast(logging.Logger, colorlog.getLogger('travo'))
        _logger.addHandler(handler)
    return _logger

def urlencode(s: str) -> str:
    """
    Encode `s` for inclusion in a URL

    EXAMPLES:

    >>> urlencode("foo/bar!")
    'foo%2Fbar%21'
    """
    return urllib.parse.urlencode({'':s})[1:]

def run(args: Sequence[str],
        check: bool = True, **kwargs: Any) -> subprocess.CompletedProcess:
    """
    A wrapper around subprocess.run

    - logs command with pretty printing
    - set check=True by default
    """
    # Backport capture_output from Python 3.6
    if kwargs.get('capture_output'):
        del kwargs['capture_output']
        kwargs['stdout'] = subprocess.PIPE
        kwargs['stderr'] = subprocess.STDOUT
    getLogger().info("Running: " + ' '.join(args))
    return subprocess.run(args, check=check, **kwargs)

def git_get_origin(cwd: str = ".") -> str:
    """
    Return the origin of the current repository

        >>> tmp_path = getfixture('tmp_path')
        >>> result = subprocess.run(["git", "init"], cwd=tmp_path)
        >>> result = subprocess.run(["git", "remote", "add", "origin", "https://xxx.yy/truc.git"], cwd=tmp_path)
        >>> git_get_origin(tmp_path)
        'https://xxx.yy/truc.git'
        >>> os.chdir(tmp_path)
        >>> git_get_origin()
        'https://xxx.yy/truc.git'
    """
    result = run(["git", "remote", "get-url", "origin"], check=False, capture_output=True, cwd=cwd)
    if result.returncode != 0:
        raise RuntimeError(result.stdout.decode().strip())
    for line in result.stdout.decode().splitlines():
        return cast(str, line)
    raise RuntimeError("Not a git repository or no origin remote configured")



"""
Utilities to write course scripts

This module implements an automatic adapter for provisionning a
Command Line Interfaces (CLI) from a Python Object Oriented Interface
defined by an object and some of its subobjects.

This is used to implement a generic main function for course scripts,
provisionning a CLI from the course object and its forge.
"""

import argparse
from argparse import ArgumentParser
import inspect
from subprocess import CalledProcessError
from typing import Callable, Optional, Any, List, Dict, cast

from .utils import getLogger
from .course import Course


# Get the value returned by inspect when a function argument has no
# default value (mypy does not like inspect._empty; it indeed is not
# clearly public)
def dummy(x: Any) -> None:
    pass


NODEFAULT = inspect.signature(dummy).parameters["x"].default


def add_parser_arguments_for_function(
    parser: ArgumentParser, function: Callable
) -> None:
    """Inspect the function signature and add the arguments to the parser

    Examples::

        >>> def f(x, y): pass
        >>> parser = ArgumentParser(prog='PROG')
        >>> add_parser_arguments_for_function(parser, f)
        >>> parser.parse_args(["a", "b"])
        Namespace(...)

    To make these examples more testable, we use pprint to display the
    result::

        >>> from pprint import pprint as pp
        >>> pp(vars(parser.parse_args(["a", "b"])))
        {'_function': <function f at ...>, 'x': 'a', 'y': 'b'}

        >>> def f(x, y=0): pass
        >>> parser = ArgumentParser(prog='PROG')
        >>> add_parser_arguments_for_function(parser, f)
        >>> pp(vars(parser.parse_args(["a", "b"])))
        {'_function': <function f at ...>, 'x': 'a', 'y': 'b'}
        >>> pp(vars(parser.parse_args(["a"])))
        {'_function': <function f at ...>, 'x': 'a', 'y': 0}

        >>> def f(force: bool = False): pass
        >>> parser = ArgumentParser(prog='PROG')
        >>> add_parser_arguments_for_function(parser, f)
        >>> pp(vars(parser.parse_args([])))
        {'--force': False, '_function': <function f at ...>}
        >>> pp(vars(parser.parse_args(["--force"])))
        {'--force': True, '_function': <function f at ...>}

        >>> def f(a, b, *args): pass
        >>> parser = ArgumentParser(prog='PROG')
        >>> add_parser_arguments_for_function(parser, f)
        >>> pp(vars(parser.parse_args(["a", "b", "c", "d", "e"])))
        {'*args': ['c', 'd', 'e'],
         '_function': <function f at ...>,
         'a': 'a',
         'b': 'b'}

    Special case: generic keyword arguments with a type of `None`:

        >>> def f(a, b, *args, **kwargs: None): pass
        >>> parser = ArgumentParser(prog='PROG')
        >>> add_parser_arguments_for_function(parser, f)
        >>> pp(vars(parser.parse_args(["a", "b", "--a", "c", "-d", "-f", "g", "--h"])))
        {'*args': ['--a', 'c', '-d', '-f', 'g', '--h'],
         '_function': <function f at ...>,
         'a': 'a',
         'b': 'b'}
    """
    parser.set_defaults(_function=function)
    signature = inspect.signature(function)
    for key, value in signature.parameters.items():
        if value.annotation == bool and not value.default:
            parser.add_argument(
                "--" + key, action="store_true", default=False, dest="--" + key
            )
        elif value.kind == value.POSITIONAL_ONLY or (
            value.kind == value.POSITIONAL_OR_KEYWORD and value.default == NODEFAULT
        ):
            parser.add_argument(key)
        elif value.kind == value.POSITIONAL_OR_KEYWORD:
            group = parser.add_mutually_exclusive_group()
            group.add_argument(key, nargs="?", default=value.default)
            group.add_argument("--" + key, default=value.default, dest=key)
        elif value.kind == value.KEYWORD_ONLY:
            parser.add_argument(
                "--" + key, default=value.default, dest="--" + key, metavar=key.upper()
            )
        elif value.kind == value.VAR_POSITIONAL:
            parser.add_argument("*" + key, nargs="*")
        elif value.kind == value.VAR_KEYWORD:
            if value.annotation is None:
                parser.prefix_chars = ""
            # else:
            #     raise NotImplementedError("Keyword arguments")


def add_subparsers_for_object_methods(
    subparsers: argparse._SubParsersAction, object: Any
) -> None:
    """
    Inspect the object, and add a subparser for all of its methods

    Return the subparsers

    Examples::

        >>> from pprint import pprint as pp

        >>> class A:
        ...     def f(self, x, y):
        ...         pass
        ...     def g(self, a, b=0):
        ...         pass

        >>> parser = ArgumentParser(prog='PROG')
        >>> subparsers = parser.add_subparsers()
        >>> add_subparsers_for_object_methods(subparsers, A())
        >>> namespace = parser.parse_args(["f", "1", "3"]); namespace
        Namespace(...)
        >>> pp(vars(namespace))
        {'_function': <bound method A.f of <travo.script.A object at ...>>,
         'x': '1',
         'y': '3'}
        >>> pp(vars(parser.parse_args(["g", "2"])))
        {'_function': <bound method A.g of <travo.script.A object at ...>>,
         'a': '2',
         'b': 0}
    """
    for name in dir(object.__class__):
        if name[0] == "_":
            continue
        method = getattr(object, name)
        if not (inspect.ismethod(method) or inspect.isfunction(method)):
            continue
        help = inspect.getdoc(method)
        if help is not None:
            help = help.splitlines()[0]
        parser = subparsers.add_parser(name, help=help, description=help)
        add_parser_arguments_for_function(parser, method)


def add_object_parser(
    parser: ArgumentParser, obj: Any, subobjects: List[Dict] = []
) -> None:
    """
    Command Line Interface from Object Oriented Interface

    This inspects the available methods of `obj`, sets up an
    :class:`ArgumentParser` mimicking it, parses the command line, and
    calls a method of `obj` accordingly. With all the benefits from
    ArgumentParser: automatic help, etc.

    For example, a command line such as:

        script foo bar truc

    will call:

        obj.foo("bar", "truc")

    Examples::

        >>> from pprint import pprint as pp
        >>> class A:
        ...     def a1(self):
        ...         pass
        ...     def a2(self, a21, a22=0):
        ...         pass
        >>> class B:
        ...     def b1(self, b11):
        ...         pass
        >>> class C:
        ...     def c1(self, c11, c12):
        ...         pass

        >>> a = A()
        >>> a.b = B()
        >>> a.b.c = C()

        >>> parser = ArgumentParser(usage="Come play with A")
        >>> add_object_parser(parser,
        ...                          a,
        ...                          subobjects=[
        ...                              dict(name="b",
        ...                                   usage='usage b',
        ...                                   subobjects=[
        ...                                       dict(name="c",
        ...                                            usage='usage c')
        ...                                   ])
        ...                              ])

        >>> parser.parse_args(["a1"])
        Namespace(...)
        >>> pp(vars(parser.parse_args(["a2", "1"])))
        {'_function': <bound method A.a2 of <travo.script.A object at ...>>,
         'a21': '1',
         'a22': 0}

        >>> pp(vars(parser.parse_args(["a2", "1", "2"])))
        {'_function': <bound method A.a2 of <travo.script.A object at ...>>,
         'a21': '1',
         'a22': '2'}
        >>> pp(vars(parser.parse_args(["b", "b1", "2"])))
        {'_function': <bound method B.b1 of <travo.script.B object at ...>>,
         'b11': '2'}
        >>> pp(vars(parser.parse_args(["b", "c", "c1", "1", "2"])))
        {'_function': <bound method C.c1 of <travo.script.C object at ...>>,
         'c11': '1',
         'c12': '2'}

        >>> parser.print_usage()
        usage: Come play with A
        >>> parser.print_help()
        usage: Come play with A
        <BLANKLINE>
        option...s:
          -h, --help  show this help message and exit
        <BLANKLINE>
        Subcommands:
          {a1,a2,b}
            a1
            a2
    """
    subparsers = parser.add_subparsers(prog=parser.prog, title="Subcommands")
    add_subparsers_for_object_methods(subparsers, obj)

    def add_subobject(name: str, subobjects: List = [], **kwargs: Any) -> None:
        subobj = getattr(obj, name)
        subparser = subparsers.add_parser(name, **kwargs)
        add_object_parser(subparser, subobj, subobjects=subobjects)

    for subobject_kwargs in subobjects:
        add_subobject(**subobject_kwargs)


def CLI(
    obj: Any,
    subobjects: List[Dict] = [],
    args: Optional[List[str]] = None,
    **kwargs: Any
) -> None:
    """
    Automatic CLI adapter

    Adapts the Python Object Oriented Interface defined by `obj` and
    some of its subobjects into a Command Line Interface.

    Examples::

        >>> class A:
        ...     def f(self, x, y):
        ...         return(x, y)
        ...     def g(self, x, y, z=1, *args, t=2):
        ...         return(x, y, z, *args, t)
        ...     def h(self, a, b, *args, **kwargs: None):
        ...         return(a, b, *args)
        >>> a = A()
        >>> a.f('x', 'y')
        ('x', 'y')
        >>> CLI(a, args=['f', 'x', 'y'])
        ('x', 'y')

        >>> a.g('x', 'y')
        ('x', 'y', 1, 2)
        >>> CLI(a, args=['g', 'x', 'y'])
        ('x', 'y', 1, 2)

        >>> a.g('x', 'y', 'z', '1', '2', t='t')
        ('x', 'y', 'z', '1', '2', 't')
        >>> CLI(a, args=['g', 'x', 'y', 'z', '1', '2', '--t', 't'])
        ('x', 'y', 'z', '1', '2', 't')

        >>> a.g('x', 'y', t='t', z='z')
        ('x', 'y', 'z', 't')
        >>> CLI(a, args=['g', 'x', 'y', '--t', 't', '--z', 'z'])
        ('x', 'y', 'z', 't')

    This one is broken:

        >>> CLI(a, args=['g', '--t', 't', '--z', 'z', 'x', 'y'])
        ('x', 'y', 1, 't')

    Checking for conflicts between positional and keyword::

        >>> a.g('x', 'y', '1', '2', z='z')
        Traceback (most recent call last):
        TypeError: g() got multiple values for argument 'z'
        >>> CLI(a, args=['g', 'x', 'y', '1', '2', '--z', 'z'])
        Traceback (most recent call last):
        ...
        SystemExit: 2

    As a special case, a generic keyword arguments with a type of
    `None` serves as a marker that all command line arguments should
    be considered as positional, regardless of whether they start with
    `-` or not.

        >>> a.h('a', 'b', 'c', '--d', 'e', '--f')
        ('a', 'b', 'c', '--d', 'e', '--f')
        >>> CLI(a, args=['h', 'a', 'b', 'c', '--d', 'e', '--f'])
        ('a', 'b', 'c', '--d', 'e', '--f')
    """
    parser = ArgumentParser(**kwargs)
    add_object_parser(parser, obj, subobjects=subobjects)

    # parser.add_argument("--loglevel", default='INFO', choices=['DEBUG', 'INFO', 'WARNING', 'CRITICAL', 'ERROR'])
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--debug",
        dest="loglevel",
        action="store_const",
        const="DEBUG",
        default="INFO",
        help="show debugging information",
    )
    group.add_argument(
        "--silent",
        dest="loglevel",
        action="store_const",
        const="WARNING",
        help="run silently",
    )

    namespace = vars(parser.parse_args(args=args))

    # Handle log level
    log = getattr(obj, "log", getLogger())
    log.setLevel(namespace["loglevel"])
    debug = namespace["loglevel"] == "DEBUG"
    del namespace["loglevel"]

    # Recover the function to call and its arguments
    if "_function" not in namespace:
        parser.print_usage()
        exit(1)

    function = namespace["_function"]
    del namespace["_function"]

    function_args = []
    function_kwargs = {}
    for key, value in namespace.items():
        if key.startswith("--"):
            function_kwargs[key[2:]] = value
        elif key.startswith("*"):
            function_args.extend(value)
        else:
            assert "--" + key not in kwargs
            # TODO this seems to fix a conflicting bug but I'm not sure what I'm doing
            function_args.append(value)
            # function_kwargs[key] = value

    if debug:
        result = function(*function_args, **function_kwargs)
    else:
        try:
            result = function(*function_args, **function_kwargs)
        except (RuntimeError, CalledProcessError) as e:
            log.error(str(e))
            # log.error("Utiliser --debug pour plus de détails")
            exit(2)

    if result is not None:
        print(result)


def main(course: Course, usage: str) -> None:
    """
    A default main function for course scripts

    It automatically adapts the Python Object Oriented Interface for
    the course and its forge into a Command Line Interface.
    """

    # Insert a helper toplevel git subcommand that pass
    # all its arguments to git
    def git(self: Course, *args: str, **kwargs: None) -> None:
        """
        Run git, passing down the forge's credentials
        """
        self.forge.git(args)

    cast(Any, Course).git = git

    CLI(
        course,
        usage=usage,
        subobjects=[
            dict(
                name="forge",
                help="Operations on the forge",
                description="Operations on the forge",
            )
        ],
    )



from dataclasses import dataclass, field
import datetime
from logging import Logger
import os
import socket
import subprocess
import time
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast
from .gitlab import anonymous_user, Forge, Job, Project, ResourceRef,\
    ResourceNotFoundError, User, Unknown, unknown
from .i18n import _


@dataclass
class Assignment:
    forge:                     Forge
    log:                       Logger
    repo_path:                 str
    instructors_path:          str
    name:                      str
    username:                  Optional[str] = None
    leader_name:               Optional[str] = None
    assignment_dir:            Optional[str] = None
    script:                    str = field(default='travo')
    jobs_enabled_for_students: bool = field(default=True)
    expires_at:                Optional[str] = None
    _repo_cache:               Optional[Project] = None

    @classmethod
    def from_url(cls,
                 url: str,
                 ) -> 'Assignment':
        ref = ResourceRef(url=url)
        ref.forge.login()
        username = ref.forge.get_current_user().username

        repo = ref.forge.get_project(ref.path)

        if repo.namespace.full_path == username:
            # The url points to the personal repo. We recover the
            # original repo by following the fork relation.
            assert repo.forked_from_project is not None
            repo = ref.forge.get_project(repo.forked_from_project.id)

        instructors_path = os.path.dirname(repo.path_with_namespace)
        name = os.path.basename(repo.path_with_namespace)

        assert username is not None

        return Assignment(forge=ref.forge,
                          log=ref.forge.log,
                          repo_path=repo.path_with_namespace,
                          name=name,
                          instructors_path=instructors_path,
                          username=username,
                          _repo_cache=repo,
                          )

    def repo(self) -> Project:
        if self._repo_cache is None:
            self._repo_cache = self.forge.get_project(self.repo_path)
        return self._repo_cache

    def get_username(self,
                     username: Optional[str] = None) -> str:
        """
        Return the user name

        `username`, if provided
        Otherwise the assignment username, if defined
        Otherwise, the login of the user on the forge
        """
        if username is not None:
            return username
        if self.username is not None:
            return self.username
        self.forge.login(anonymous_ok=True)
        return cast(str, self.forge.get_current_user().username)

    ##########################################################################
    # Methods meant to be overriden for configuration

    def personal_repo_path(self,
                           username: Optional[str] = None
                           ) -> str:
        """
        Return the path on the forge of the student's personal repository for this assignment

        Example:

            >>> assignment = getfixture("standalone_assignment")
            >>> assignment.personal_repo_path()
            'travo-test-etu/TestAssignment-2...'
        """
        username = self.get_username(username)
        return username+"/"+os.path.basename(self.repo_path)

    def personal_repo_name(self) -> str:
        """
        Return the name of the student's personal repository for the given assignment

        This method may typically be overriden by the course.

        Example:

            >>> assignment = getfixture("standalone_assignment")
            >>> assignment.personal_repo_name()
            'Test assignment - 2...'
        """
        return self.repo().name

    def submissions_forked_from_path(self) -> Union[str, Unknown]:
        """Return the path of the repository that submissions should be a fork of.

        By default, submissions are direct forks of the assignment
        repository (subclasses may have different rules). If
        leader_name is specified, then submissions are forks of the
        leader's submission which should exist.
        """
        if self.leader_name is not None:
            return self.personal_repo_path(self.leader_name)
        return self.repo_path

    def submissions_forked_from_missing(self) -> None:
        assert False

    def submissions_search_from(self) -> Tuple[Project, int]:
        """Return a project `p` and an int `d` such that the submissions for
        this assignment are all the forks of `p` of depth `d`
        """
        return (self.repo(), 1)

    def get_submission_username(self, project: Project) -> Optional[str]:
        """Return the username for the given submission

        This method is also used in the folloging use case: let S be a
        submission which is a fork of this project. Is this project the
        submission of the leader, or the assignment? To this end, this
        method should return None in the latter case.

        Example:

            >>> assignment = getfixture("standalone_assignment")
            >>> assignment.get_submission_username(assignment.repo())
            >>> project = getfixture("standalone_assignment_submission")
            >>> assignment.get_submission_username(project)
            'travo-test-etu'
        """
        if project.path_with_namespace == self.repo_path:
            return None
        else:
            return project.get_creator().username


    ##########################################################################

    def personal_repo(self) -> Project:
        return self.forge.get_project(self.personal_repo_path())

    def has_personal_repo(self) -> bool:
        """
        Return whether the user already has a personal repository for this assignment

        Examples::

            >>> assignment = getfixture("standalone_assignment")
            >>> assignment.has_personal_repo()
            False
            >>> my_repo = assignment.ensure_personal_repo()
            >>> assignment.has_personal_repo()
            True
            >>> assignment.remove_personal_repo(force=True)
            >>> assignment.has_personal_repo()
            False
            """
        try:
            self.personal_repo()
            return True
        except ResourceNotFoundError:
            return False

    def ensure_personal_repo(self,
                             leader_name: Optional[str] = None) -> Project:
        """
        Return the personal repository for this assignment

        Creating it and configuring it if needed.

        Example::

            >>> assignment = getfixture("standalone_assignment")
            >>> forge = assignment.forge
            >>> repo = assignment.repo()

            >>> my_repo = assignment.ensure_personal_repo()
            >>> my_repo.path
            'TestAssignment-20...'
            >>> my_repo.name
            'Test assignment - 20...'
            >>> assert my_repo.forked_from_project.id == repo.id

            >>> my_repo = assignment.ensure_personal_repo()
            >>> share, = my_repo.shared_with_groups
            >>> assert share['group_full_path'] == assignment.instructors_path

        Tear down::

            >>> assignment.remove_personal_repo(force=True)
        """
        # travo_raise_info "- Vérification que le devoir a un dépôt personnel configuré"
        repo = self.repo()
        instructors = self.forge.get_group(self.instructors_path)

        my_repo = repo.ensure_fork(
            path=self.personal_repo_path(),
            name=self.personal_repo_name(),
            forked_from_path=self.submissions_forked_from_path(),
            forked_from_missing=self.submissions_forked_from_missing,
            visibility='private',
            emails_disabled=True,
            default_branch=repo.default_branch,
            jobs_enabled=self.jobs_enabled_for_students,
        )
        if my_repo.default_branch == "null":
            web_url = my_repo.web_url
            self.log.error(f"""Dépôt personnel corrompu: {web_url}
Consultez le à l'adresse ci-dessus; s'il est vide, vous
pouvez le détruire avec la commande:

    {self.script} forge_remove_project {my_repo.path_with_namespace}

Il sera reconstruit lors du prochain dépôt""")

        if not any(share['group_id'] == instructors.id
                   for share in my_repo.shared_with_groups):
            self.log.info(f"- Configuration accès enseignants "
                          f"(groupe {instructors.path})")
            try:
                my_repo.share_with(instructors,
                                   access=my_repo.AccessLevels.MAINTAINER,
                                   expires_at=self.expires_at)
            except KeyError:
                member_ids = [user['id'] for user in my_repo.get_members()]
                for instructor in instructors.get_members():
                    if instructor.id not in member_ids:
                        my_repo.share_with(instructor,
                                           access=my_repo.AccessLevels.MAINTAINER,
                                           expires_at=self.expires_at)

        self.log.debug("- Configuration badge:")
        artifacts_url = self.forge.base_url+'/%{project_path}/-/jobs/artifacts/%{default_branch}/'
        link_url = artifacts_url + 'file/feedback/scores.html?job=autograde'
        image_url = artifacts_url + 'raw/feedback/scores.svg?job=autograde'

        my_repo.ensure_badge(name="Scores",
                             link_url=link_url,
                             image_url=image_url)

        self.log.info("- Votre dépôt personnel:")
        self.log.info(f"  {my_repo.web_url}")
        return my_repo

    def remove_personal_repo(self, force: bool=False) -> None:
        """
        Remove the users' personal repository for this assignment
        """
        self.forge.remove_project(self.personal_repo_path(), force=force)

    def ensure_clone_configuration(self,
                                   assignment_dir: str,
                                   ) -> None:
        """
        Ensure that the clone of the assignment is configured

        - 'origin' points to the personal repository for the assignment
        - 'user.name' and 'user.mail' are set

        This assumes that the given clone exists.

        Examples:

        We assume that the student has a clone of the assignment
        in his work directory for the course::

            >>> import subprocess
            >>> assignment      = getfixture("standalone_assignment")
            >>> assignment_dir  = getfixture("standalone_assignment_dir")
            >>> forge = assignment.forge
            >>> repo = assignment.repo()
            >>> print(forge.git(["clone", repo.http_url_to_repo, assignment_dir],
            ...                  stdout=subprocess.PIPE,
            ...                  stderr=subprocess.STDOUT).stdout.decode())
            Clon... .../Assignment-20...

        After running this command::

            >>> assignment.ensure_clone_configuration(assignment_dir)

        the student is guaranteed to have a personal repository for
        the assignment::

            >>> path = assignment.personal_repo_path()
            >>> project = forge.get_project(path)
            >>> project.path_with_namespace
            'travo-test-etu/TestAssignment-20...'
            >>> print(forge.git(["config", "--local", "user.name"],
            ...                 capture_output=True,
            ...                 cwd=assignment_dir).stdout.decode(), end='')
            Étudiant de test pour travo
            >>> print(forge.git(["config", "--local", "user.email"],
            ...                 capture_output=True,
            ...                 cwd=assignment_dir).stdout.decode(), end='')
            pierre.froidevaux@viacesi.fr

        And the remote of the clone will be set appropriately:

            >>> print(forge.git(["remote", "-v"],
            ...           capture_output=True,
            ...           cwd=assignment_dir,
            ...       ).stdout.decode(), end='')
            origin	https://gitlab-test.info.uqam.ca/travo-test-etu/TestAssignment-20....git (fetch)
            origin	https://gitlab-test.info.uqam.ca/travo-test-etu/TestAssignment-20....git (push)

        This is an idempotent operation:

            >>> assignment.ensure_clone_configuration(assignment_dir)

        Tear down::

            >>> assignment.remove_personal_repo(force=True)
        """
        self.forge.ensure_local_git_configuration(assignment_dir)

        self.log.debug("Ensure that the remote of this repository is configured")
        project = self.ensure_personal_repo()
        result = self.forge.git(["remote", "-v"],
                                cwd=assignment_dir,
                                capture_output=True)
        if project.http_url_to_repo in result.stdout.decode():
            self.log.debug("Personal remote already configured")
        else:
            self.log.info(f"Reconfiguration du dépôt distant de la copie de travail: "
                          f"{project.http_url_to_repo}")
            self.forge.git(["remote", "set-url", "origin", project.http_url_to_repo],
                           cwd=assignment_dir)

    def assignment_clone(self, assignment_dir: str) -> None:
        """
        Create the student's local clone of the assignment

        - from their personal repo if it exists
        - from the original repo otherwise

        If the user is logged in as `anonymous`, then the
        original repo is cloned. It needs to be public.
        """
        self.log.info(f"Création d'une copie de travail du devoir {self.name}"
                      f"dans {assignment_dir}")

        user = self.forge.get_current_user()
        source: Optional[Project] = None
        if user is not anonymous_user:
            try:
                source = self.personal_repo()
            except ResourceNotFoundError:
                pass
            else:
                # Support for accidental empty personal repository
                if not source.get_branches():
                    source = None

        if source is None:
            source = self.repo()
            self.log.info(f"  à partir du dépôt d'origine {source.http_url_to_repo}")
        else:
            self.log.info(f"  à partir de votre dépot personnel "
                          f"{source.http_url_to_repo}")

        self.forge.git(["clone", source.http_url_to_repo, assignment_dir],
                       anonymous_ok=True)

    def check_assignment_dir(self, assignment_dir: str) -> None:
        """
        Check the assignment directory

        Raise an error if it does not exist or is dubious
        """
        self.log.info(f"- Vérification de la présence du devoir {assignment_dir}")
        if not os.path.exists(assignment_dir):
            self.log.error(f"""Copie de travail {assignment_dir} non trouvée.
Vous pouvez la télécharger avec:

    {self.script} fetch {assignment_dir}
""")
            raise RuntimeError("missing assignment directory")
        if not os.path.isdir(assignment_dir):
            self.log.error(f"""Un fichier ou répertoire {assignment_dir} existe
mais n'est pas une copie de travail du devoir.
Déplacez ou supprimez le
""")
            raise RuntimeError("corrupted assignment directory")

    def fetch_branch(self,
                     assignment_dir: str,
                     branch: Optional[str],
                     content: str,
                     on_failure: str = 'error',
                     ) -> None:
        """
        Fetch from a branch of the assignment repository

        If branch_name is None, the default branch is used.

        This will try to merge the branch `branch` of the assignment
        repository (if it exists) into the local clone. In case of
        merge conflict, the local clone is rolled back to its original
        state, and a warning or error is emmitted depending on the
        value of `on_failure`.


        Assumption: `assignment_dir` is a clone of the assignment
        repository, with a local git configuration (see
        :ref:`Gitlab.ensure_local_git_configuration`).

        """
        self.log.info(f"Intégration des {content} depuis le dépot d'origine")

        source = self.repo()
        if branch is None:
            branch = source.default_branch
        if not any(b['name'] == branch
                   for b in source.get_branches()):
            self.log.info(f"Pas de {content} pour ce devoir")
            return

        # self.check_assignment("$assignment")
        # Share this with fetch
        def git(args: List[str], **kwargs: Any) -> subprocess.CompletedProcess:
            return self.forge.git(args, cwd=assignment_dir, **kwargs)

        self.log.info("- Sauvegarde préalable:")
        git(["commit", "--all", "-m", f"Sauvegarde avant intégration des {content}"],
            check=False)
        self.log.info("- Téléchargement:")
        git(["fetch", source.http_url_to_repo, branch])
        # travo_raise_error f"Échec au téléchargement des {content}"
        self.log.info("- Tentative d'intégration:")
        if git(["merge", "-m", f"Intégration des {content}", "FETCH_HEAD"],
               check=False).returncode != 0:
            git(["merge", "--abort"])
            message = _("abandon failing merge", content=content)
            if on_failure == 'warning':
                self.log.warning(message)
            else:
                assert on_failure == 'error'
                raise RuntimeError(message)

    def fetch(self, assignment_dir: Optional[str] = None,
              force: bool = False) -> None:
        """
        fetch the given assignment

        The user must have an account on the forge:
        - to check whether the user has a personal repository for this assignment
        - to setup the local git configuration

        If logged in as `anonymous`, fetch will ignore any personnal
        repository the user may have, will skip the git configuration
        and won't fetch updates or erratas. We could try harder on
        this last point if there is a use case for it.
        """
        self.forge.login(anonymous_ok=True)
        user = self.forge.get_current_user()
        def git(args: List[str], **kwargs: Any) -> subprocess.CompletedProcess:
            return self.forge.git(args, cwd=assignment_dir, **kwargs)

        # check_assignment "$assignment"
        if assignment_dir is None:
            assignment_dir = self.assignment_dir
            assert assignment_dir is not None

        if not os.path.exists(assignment_dir):
            self.assignment_clone(assignment_dir)
        elif user is not anonymous_user and self.has_personal_repo():
            personal_repo = self.forge.get_project(self.personal_repo_path())
            self.ensure_clone_configuration(assignment_dir)
            self.log.info("Intégration des mises à jour depuis votre dépot personnel")
            self.log.info("- Sauvegarde préalable:")
            git(["commit", "--all", "-m", "Sauvegarde avant mise à jour"], check=False),
            self.log.info("- Téléchargement:")
            # TODO: nice error message in case of failure
            # - Échec au téléchargement des mises à jour"
            git(["fetch", personal_repo.http_url_to_repo])
            self.log.info("- Tentative d'intégration:")

            if git(["merge", "-m", "Intégration des mises à jour", "FETCH_HEAD"],
                   check=False).returncode != 0:
                git(["merge", "--abort"])
                if force:
                    now = datetime.datetime.now().isoformat(sep="_", timespec="minutes")
                    # Something special needs to be done for "."
                    assert assignment_dir != "."
                    backup_dir = f"{assignment_dir}_{now}"
                    self.log.warning(f"""La mise à jour n'a pas pu se faire du fait d'un conflit
  Votre devoir local va être renommé en {backup_dir}
  et une copie fraîche du sujet téléchargée à la place
""")
                    os.rename(assignment_dir, backup_dir)
                    self.fetch(assignment_dir)
                else:
                    self.log.error("- La mise à jour n'a pas pu se faire du fait d'un conflit.")
                    self.log.info(f"""  Pour renommer votre devoir local et forcer la mise à jour,
  utilisez l'option force; en ligne de commande:

       {self.script} fetch ... --force
""")
                    raise RuntimeError(_("fetch failed conflict"))
        if user is not anonymous_user:
            self.forge.ensure_local_git_configuration(assignment_dir)
            # fetch_branch may require a commit which require git to be configured
            # we could try harder if there is a use case for it
            self.fetch_branch(assignment_dir, branch=None, content=_("updates"))
            self.fetch_branch(assignment_dir, branch="errata", content=_("erratas"), on_failure='warning')

    def submit(self,
               assignment_dir: Optional[str] = None,
               leader_name: Optional[str] = None
               ) -> None:
        """
        submit the given assignment
        """
        self.forge.login()
        if assignment_dir is None:
            assignment_dir = self.assignment_dir
            assert assignment_dir is not None

        self.log.info(f"Soumission de {assignment_dir}:")

        self.check_assignment_dir(assignment_dir)

        # in case of failure, raise: "Échec de la soumission"
        # Must be a failure connecting to gitlab
        user = self.forge.get_current_user()
        assert isinstance(user, User)

        personal_repo = self.ensure_personal_repo(leader_name=leader_name)
        url = personal_repo.http_url_to_repo

        self.ensure_clone_configuration(assignment_dir)

        hostname = socket.gethostname()

        def git(args: List[str], **kwargs: Any) -> Any:
            return self.forge.git(args, cwd=assignment_dir, **kwargs)

        self.log.info("- Enregistrement des changements:")
        if git(["commit", "--all", "-m", f"Soumission depuis {hostname} par {user.name}"],
               check=False).returncode != 0:
            self.log.info("  Pas de changement à enregistrer")

        self.log.info("- Envoi des changements:")
        # With a newly created repo personal_repo.default_branch may
        # still be master even if the assignment repo is using,
        # e.g. main. Force using the assignment repo's default branch.
        branch = self.repo().default_branch
        if git(["push", url, branch],
               check=False).returncode != 0:
            raise RuntimeError(_("submission failed"))
        # Force an update of origin/master (or whichever the origin default branch is)
        git(["update-ref", f"refs/remotes/origin/{branch}", branch])

        self.log.info(f"""- Soumission effectuée. Vous pouvez consulter votre dépôt:
    {url}
""")

    def submission(self) -> "Submission":
        """
        Return the user's submission for this assignment

        Fails if the user has no submission
        """
        if not self.has_personal_repo():
            raise RuntimeError(
                f"Dépôt personnel inexistant sur GitLab\n"
                f"Merci de déposer `{self.name}` (avec submit)")
        project = self.personal_repo()
        return Submission(project, assignment=self)

    def submissions(self) -> List["Submission"]:
        """
        Return all the submissions for this assignment

        The submissions are found among the forks or forks of forks.
        To distinguish between actual submissions and,
        e.g. intermediate forks used to model student groups, the
        current criteria is whether the fork resides in a user
        namespace, rather than a group namespace. The criteria used to
        be whether the fork had an owner.
        """
        # This is an instructor feature; leader_name is a student feature
        assert self.leader_name is None
        repo, fork_depth = self.submissions_search_from()
        forks = repo.get_forks(recursive=fork_depth, simple=True)
        return [Submission(fork, assignment=self)
                for fork in forks
                if fork.namespace.kind == 'user']

    def status(self) -> "SubmissionStatus":
        """
        Return the status of this assignment for its user
        """
        if self.has_personal_repo():
            result = self.submission().status()
        else:
            try:
                self.repo()
                status = "released"
            except ResourceNotFoundError:
                status = "not released"
            result = SubmissionStatus(student=self.get_username(),
                                      status=status,
                                      autograde_status="none")
        return result

    def collect_status(self) -> List["SubmissionStatus"]:
        """
        Return the status of all the submissions for this assignment
        """
        return [submission.status()
                for submission in self.submissions()]


# Only consider running or finished jobs
# this excludes: 'canceled', 'skipped', 'manual',
job_status = ['failed', 'created', 'pending', 'running', 'success']


def job_status_priority(job: dict) -> Tuple[int, str]:
    return job_status.index(job['status']), job['created_at']


@dataclass
class SubmissionStatus:
    student: str
    status: str  # Literal['not released', 'released', 'submitted', 'autograded']
    autograde_status: str
    leader_name: Optional[str] = None
    team: Optional[Dict[str, Project]] = None
    autograde_job: Optional[Job] = field(default=None, repr=False)
    submission: Optional["Submission"] = field(default=None, repr=False)

    def is_submitted(self) -> bool:
        return self.status in ["submitted", "autograded"]


@dataclass
class Submission:
    repo: Project
    assignment: Assignment
    student: str = field(default=cast(str, None))

    def __post_init__(self) -> None:
        student = self.repo.gitlab.get_user(self.repo.creator_id).username
        assert student is not None
        self.student = student

    def get_autograde_job(self) -> Optional[Job]:
        """
        Return autograde job

        This selects the jobs for the latest commit of the default
        branch, which are running or finished. If there is none,
        return None. Otherwise, return that which is the most advanced
        in the following order: "success" > "failure" > "running" >
        "created" > "failed", breaking ties with creation time.
        """
        # TODO remove old feedback
        # TODO deadline management:
        # - lookup appropriate commit instead of the latest as above
        #   https://docs.gitlab.com/ee/api/commits.html avec option "until"

        if not self.assignment.jobs_enabled_for_students:
            return None

        repo = self.repo
        try:
            branch = repo.get_branch(repo.default_branch)
        except Exception:  # would be nicer with a ResourceNotFoundError
            # This can actually fail with an empty repository
            return None

        commit_id = branch['commit']['id']
        forge = self.repo.gitlab

        jobs = forge.get_json(f"/projects/{repo.id}/jobs",
                              data={'scope': job_status},
                              depaginate=True)
        jobs = [job
                for job in jobs
                if job['commit']['id'] == commit_id
                and job['status'] not in ['canceled', 'skipped', 'manual']
                ]

        if jobs:
            job = max(jobs, key=job_status_priority)
            return job

        return None

    def ensure_autograded(self,
                          force_autograde: bool = False) -> Job:
        """
        Ensure that this submission was autograded by CI and feedback is available

        If `force_autograde` is set, a new autograde is forced.

        Return the job that computed the autograde
        """
        repo = self.repo
        forge = repo.gitlab
        if force_autograde:
            job = None
        else:
            status = self.status()
            job = status.autograde_job
            if status.status == 'autograded':
                return job
            if job is not None:
                url = job['web_url']
                if status.autograde_status == "success_no_artifact":
                    raise RuntimeError(
                        f"Correction automatique terminée mais sans artefact: {url}")
                forge.log.info(
                    f"En attente de la correction automatique: {url}")

        if job is None:
            # Run a new pipeline
            # TODO: could trigger directly a job instead of a pipeline
            pipeline = forge.post(f"/projects/{repo.id}/pipeline",
                                  data={"ref": repo.default_branch}
                                  ).json()
            message = pipeline.get('message')
            if message == '403 Forbidden':
                raise RuntimeError(
                    f"Accès «Maintainer» manquant: {self.repo.web_url}")
            if message is not None:
                raise RuntimeError(
                    f"Échec au lancement de la correction automatique ({message}): "
                    f"{repo.web_url}")
            forge.log.info(
                f"Lancement d'une nouvelle correction automatique: {pipeline['web_url']}")

        # Wait for autograde execution
        import tqdm  # type: ignore
        progress_bar = tqdm.tqdm(
            desc='',
            bar_format="En attente de la correction automatique ({desc}): {elapsed}")
        progress_bar.display()

        if job is None:
            while pipeline['status'] in ['created', 'pending', 'running']:
                progress_bar.desc = pipeline['status']
                progress_bar.update()
                time.sleep(1)
                pipeline = forge.get_json(f"/projects/{repo.id}/pipelines/{pipeline['id']}")

            job, = forge.get_json(
                f"/projects/{repo.id}/pipelines/{pipeline['id']}/jobs")

        while job['status'] in ['created', 'pending', 'running']:
            progress_bar.desc = job['status']
            progress_bar.update()
            time.sleep(1)
            job = forge.get_json(f"/projects/{repo.id}/jobs/{job['id']}")

        progress_bar.close()

        # Check job status
        if job['status'] != 'success':
            forge.log.warning(f"Failed job: {job['web_url']}")
            raise RuntimeError("Échec de la correction automatique")
        if 'artifacts_file' not in job:
            forge.log.warning(f"Job: {job['web_url']}")
            raise RuntimeError("Correction automatique terminée mais sans artefact")
        return job

    def force_autograde(self) -> Job:
        """
        Force the autograde and feedback of the submission by CI.
        Does not wait for the end of the job.

        Return the job that is computing the autograde.
        """
        repo = self.repo
        forge = repo.gitlab
        # Run a new pipeline
        # TODO: could trigger directly a job instead of a pipeline
        pipeline = forge.post(f"/projects/{repo.id}/pipeline",
                              data={"ref": repo.default_branch}
                              ).json()
        message = pipeline.get('message')
        if message == '403 Forbidden':
            raise RuntimeError(
                f"Accès «Maintainer» manquant: {self.repo.web_url}")
        if message is not None:
            raise RuntimeError(
                f"Échec au lancement de la correction automatique ({message}): "
                f"{repo.web_url}")
        forge.log.info(
            f"Lancement d'une nouvelle correction automatique: {pipeline['web_url']}")
        job, = forge.get_json(
            f"/projects/{repo.id}/pipelines/{pipeline['id']}/jobs")
        return job

    def get_leader_and_team(self) -> Tuple[str, Dict[str, Project]]:
        """
        Return the leader of the submission and the repositories of the team mates

        Output: a tuple whose first entry is the username of the leader and the
        second entry is a dictionary mapping usernames of team mates to their
        submissions.
        """
        repo = self.repo
        forked_from_project = self.repo.forked_from_project
        if forked_from_project is None:
            raise ValueError("Team information not available "
                             "for a project with no forked_from information")
        leader_name = self.assignment.get_submission_username(forked_from_project)
        if leader_name is not None:
            repo = forked_from_project
        else:
            leader_name = self.assignment.get_submission_username(repo)
            assert leader_name is not None

        team = {leader_name: repo}
        if repo.forks_count > 0:
            for fork in repo.get_forks(simple=True):
                username = self.assignment.get_submission_username(fork)
                assert username is not None
                team[username] = fork
        return (leader_name, team)

    def status(self) -> SubmissionStatus:
        status = "submitted"

        leader_name, team = self.get_leader_and_team()

        autograde_job = self.get_autograde_job()
        if autograde_job is not None:
            autograde_status = cast(str, autograde_job['status'])
            if not any(artifact["filename"] == "artifacts.zip"
                       for artifact in autograde_job.get("artifacts", [])):
                autograde_status = "success_no_artifact"
            if autograde_status == "success":
                status = "autograded"
        else:
            autograde_status = "none"

        return SubmissionStatus(
            student=self.student,
            leader_name=leader_name,
            team=team,
            autograde_status=autograde_status,
            autograde_job=autograde_job,
            status=status,
            submission=self,
            )



"""
Implementation of the console scripts for travo

- travo
- travo_echo_travo_token
"""

import os
from typing import Any, Optional
from travo import Assignment
from travo.script import CLI
from travo.utils import git_get_origin
from travo import Homework


class Travo:
    """
    This class defines the command line interface for the travo script
    """

    def info(
        self,
        url: str = ".",
        fixup: bool = False,
        group: Optional[str] = None,
        copy: Optional[str] = None,
    ) -> None:
        """
        Get, check and print information on the repository.

        The repository can be either the instructor's assignment or a student submission.
        If the former case, the information is iterated on all student copies.

        --group indicates the correction group to check (if any)

        --fixup tries to fix configuration mismatch (visibility, etc.)

        --copy indicate to work on the given copy (and not all)
        """

        homework = Homework(url)
        if group is not None:
            homework.get_group(group)
        if copy is None:
            forks = homework.get_copies()
        else:
            homework.assignment = homework.project  # assume assigment
            forks = [homework.get_project(copy)]

        for fork in forks:
            homework.print_info(fork, fixup=fixup)

    def search_forks(
        self,
        url: str = ".",
        fixup: bool = False,
        deep: bool = False,
        group: Optional[str] = None,
    ) -> None:
        """
        Search for possible missing forks [instructor]

        For some reason, the fork relationship can be lost with gitlab, for instance
        the `fork` button was not used or fork was made private.

        Note: the search of forks can be slow.

        --deep indicates to search among more potential projects.

        --group indicates the correction group to check (if any)

        --fixup tries to fix configuration mismatch (visibility, etc.)
        """

        homework = Homework(url)
        homework.assignment = homework.project  # assume assigment
        if group is not None:
            homework.get_group(group)

        forks = homework.project.get_possible_forks(deep=deep, progress=True)
        for fork in forks:
            homework.print_info(fork, fixup=fixup)

    def collect(self, url: str = ".", dir: str = "forks") -> None:
        """
        Collect the student repositories [instructor]

        Either a single copy, or all the students' copies if the instructor's assignment is used.

        --dir is the target directory.
        """
        # TODO: merge with Assignment.collect_forks
        homework = Homework(url)
        forks = homework.get_copies()
        template = "{user}-{id}"

        for fork in forks:
            if fork.owner is None:
                continue
            path = os.path.join(
                dir, template.format(user=fork.owner.username, id=fork.id)
            )
            homework.print_info(fork)
            fork.clone_or_pull(path)

    def fetch(self, url: str = ".", assignment_dir: Optional[str] = None) -> None:
        """
        Fetch the assignment [student]

        Fetch assignment from URL, optionally specifying an assignment_dir:

            Travo.fetch(url)
            Travo.fetch(url, assignment_dir)

        Update an already fetched assignment:

            Travo.fetch(assignment_dir)
        """
        if not url.startswith("https:"):
            assert assignment_dir is None
            assignment_dir = url
            url = git_get_origin(assignment_dir)

        assignment = Assignment.from_url(url)

        if assignment_dir is None:
            assignment_dir = os.path.basename(assignment.repo_path)

        assignment.fetch(assignment_dir)

    def submit(self, assignment_dir: str = ".") -> None:
        """
        Submit the copy [student]
        """

        url = git_get_origin(assignment_dir)
        assignment = Assignment.from_url(url)
        assignment.submit(assignment_dir)

    @staticmethod
    def formgrader(assignment: Optional[str] = None, in_notebook: bool = False) -> Any:
        """
        Launch nbgrader's formgrader
        """
        from travo.jupyter_course import JupyterCourse

        return JupyterCourse.formgrader(assignment, in_notebook)

    @staticmethod
    def validate(*files: str) -> None:
        """
        Launch nbgrader's validate
        """
        from travo.jupyter_course import JupyterCourse

        return JupyterCourse.validate(*files)


def test_travo(standalone_assignment: Assignment, tmp_path: str) -> None:
    url = standalone_assignment.repo().http_url_to_repo
    assignment_dir = os.path.join(tmp_path, "Assignment")

    travo = Travo()
    travo.fetch(url, assignment_dir)

    travo.fetch(assignment_dir)
    travo.submit(assignment_dir)

    # Tear down
    standalone_assignment.remove_personal_repo()


usage = """travo [fetch|submit] ...

For students

Fetch the latest version of the assignment:

    travo fetch <url> <dir>
    travo fetch <url>
    travo fetch <dir>

Submit the assignment:

    travo submit <dir>

where `<url>` is the url of the Git repository holding the
assignment, and <dir> is the (to be created) local working copy
for the assignment. By default, the working copy is created in a
subdirectory of current working directory with basename matching
that of the assignment.

More help:

    travo --help
"""


def travo() -> None:
    """
    Entrypoint for the main `travo` console script
    """
    CLI(Travo(), usage=usage)


def travo_echo_travo_token() -> None:
    """
    Entrypoint for the `travo-echo-travo-token console script

    This script is used as GIT_ASKPASS callback to provide the gitlab
    authentication token to git
    """
    print(os.environ["TRAVO_TOKEN"])



import datetime
import os.path
import pytest  # type: ignore
import random
import string
from typing import Iterator

from travo.gitlab import GitLab, GitLabTest, Group, Project, User
from travo.course import Course
from travo.assignment import Assignment


@pytest.fixture
def gitlab() -> GitLab:
    forge = GitLabTest()
    forge.log.setLevel("DEBUG")
    return forge


@pytest.fixture
def test_run_id() -> str:
    """
    A (statistically) unique id for this test run
    """
    now = datetime.datetime.now().isoformat().replace(":", "-")
    return (
        now + "-" + "".join(random.choices(string.ascii_uppercase + string.digits, k=5))
    )


@pytest.fixture
def project_path(test_run_id: str) -> str:
    return f"travo-test-etu/temporary-test-projet-{test_run_id}"


@pytest.fixture
def project_name(test_run_id: str) -> str:
    return f"temporary test projet created at {test_run_id}"


@pytest.fixture
def project(gitlab: GitLab, project_path: str, project_name: str) -> Iterator[Project]:
    project = gitlab.ensure_project(project_path, project_name)
    # TODO: factor out duplication with assignment_repo
    project.ensure_file("README.md")
    yield project
    gitlab.remove_project(project_path, force=True)


@pytest.fixture
def group_path(test_run_id: str) -> str:
    return f"temporary-test-group-{test_run_id}"


@pytest.fixture
def group_name(test_run_id: str) -> str:
    return f"temporary test group created at {test_run_id}"


@pytest.fixture
def group(gitlab: GitLab, group_path: str, group_name: str) -> Iterator[Group]:
    yield gitlab.ensure_group(group_path, group_name)
    gitlab.remove_group(group_path, force=True)

@pytest.fixture
def user_name() -> str:
    return "travo-test-etu"


@pytest.fixture
def user(gitlab: GitLab, user_name: str) -> User:
    return gitlab.get_user(user_name)


@pytest.fixture
def other_user(gitlab: GitLab) -> User:
    return gitlab.get_user("blondin_al")


@pytest.fixture
def fork_path(user_name: str, test_run_id: str) -> str:
    return f"{user_name}/temporary-forked-project-{test_run_id}"


@pytest.fixture
def fork_name(test_run_id: str) -> str:
    return f"temporary forked project created at {test_run_id}"


@pytest.fixture
def standalone_assignment_namespace(gitlab: GitLab) -> Group:
    return gitlab.ensure_group(path="TestGroup",
                               name="Test Group",
                               with_projects=False,
                               visibility="public",
                               )


@pytest.fixture
def standalone_assignment(gitlab: GitLab,
                          user_name: str,
                          standalone_assignment_namespace: Group,
                          test_run_id: str) -> Iterator[Assignment]:
    repo = gitlab.ensure_project(
        f"TestGroup/TestAssignment-{test_run_id}", f"Test assignment - {test_run_id}")
    repo.ensure_file(
        "README.md",
        branch="master")
    yield Assignment(forge=gitlab,
                     log=gitlab.log,
                     repo_path=repo.path_with_namespace,
                     name=repo.path,
                     instructors_path="TestGroup",
                     username=user_name,
                     )
    gitlab.remove_project(repo.path_with_namespace, force=True)


@pytest.fixture
def standalone_assignment_dir(tmp_path: str, test_run_id: str) -> str:
    return os.path.join(tmp_path, f"Assignment-{test_run_id}")


@pytest.fixture
def standalone_assignment_submission(
    standalone_assignment: Assignment
) -> Iterator[Project]:
    project = standalone_assignment.ensure_personal_repo()
    yield project
    standalone_assignment.forge.remove_project(project.path_with_namespace, force=True)


@pytest.fixture
def course() -> Course:
    return Course(name="Test course",
                  forge=GitLabTest(),
                  path="TestCourse",
                  student_dir="~/TestCourse",
                  assignments_group_path="TestCourse/2020-2021"
                  )
    # course.log.setLevel("DEBUG")

@pytest.fixture
def rich_course() -> Course:
    return Course(name="Test course",
                  forge=GitLabTest(),
                  path="TestCourse",
                  session_path="2020-2021",
                  assignments=["Assignment1", "Assignment2"],
                  student_groups=["Group1", "Group2"],
                  student_dir="~/TestCourse",
                  )

@pytest.fixture
def course_assignment_group(course: Course) -> Group:
    course.forge.ensure_group(path=course.path, name=course.name)
    return course.forge.ensure_group(
        path=course.assignments_group_path, name=course.assignments_group_name
    )


@pytest.fixture
def assignment_path(test_run_id: str) -> str:
    return f"temporary-assignment-{test_run_id}"


@pytest.fixture
def assignment_name(test_run_id: str) -> str:
    return f"temporary assignment created at {test_run_id}"


@pytest.fixture
def assignment_repo(
    course: Course,
    course_assignment_group: Group,
    assignment_path: str,
    assignment_name: str,
) -> Iterator[Project]:
    path = course.assignment_repo_path(assignment_path)
    project = course.forge.ensure_project(path, assignment_name)
    project.ensure_file(
        "README.md",
        branch="master")
    yield project
    course.forge.remove_project(path, force=True)


@pytest.fixture
def assignment_personal_repo(
    course: Course, assignment_path: str, assignment_repo: Project
) -> Iterator[Project]:
    project = course.assignment(assignment_path).ensure_personal_repo()
    yield project
    course.forge.remove_project(project.path_with_namespace, force=True)


@pytest.fixture
def student_work_dir(course: Course) -> str:
    return course.ensure_work_dir()



from typing import Any, Dict, List, Optional, Tuple
from nbgrader.api import Gradebook, MissingEntry, InvalidEntry  # type: ignore
from .utils import getLogger
from sqlalchemy.orm.exc import FlushError  # type: ignore
from sqlalchemy.exc import IntegrityError, StatementError  # type: ignore
from tornado.log import app_log


def to_args(item: Any, positional: List['str']) -> Tuple[List['str'], Dict['str', Any]]:
    kwargs = dict(item.to_dict())
    args = [kwargs[key] for key in positional]
    for key in positional:
        del kwargs[key]
    return args, kwargs


def merge_assignment_gradebook(source: Gradebook, target: Gradebook) -> None:
    """
    Merge the gradebook's assignment information into the target gradebook

    Grades and submission informations are ignored

    Assumption: there is a single assignment
    """
    assignment, = source.assignments
    target.update_or_create_assignment(assignment.name)

    for notebook in assignment.notebooks:
        args, kwargs = to_args(notebook, ['name'])
        target.update_or_create_notebook(*args, assignment.name, **kwargs)
        for cell in notebook.grade_cells:
            args, kwargs = to_args(cell, ['name', 'notebook', 'assignment'])
            target.update_or_create_grade_cell(*args, **kwargs)
        for cell in notebook.task_cells:
            args, kwargs = to_args(cell, ['name', 'notebook', 'assignment'])
            target.update_or_create_task_cell(*args, cell_type=cell.cell_type, **kwargs)
        for cell in notebook.source_cells:
            args, kwargs = to_args(cell, ['name', 'notebook', 'assignment'])
            target.update_or_create_source_cell(*args, **kwargs)
        for cell in notebook.solution_cells:
            args, kwargs = to_args(cell, ['name', 'notebook', 'assignment'])
            target.update_or_create_solution_cell(*args, **kwargs)


def merge_submission_gradebook(source: Gradebook,
                               target: Gradebook,
                               back : bool = False,
                               on_inconsistency: str = "ERROR",
                               # Literal("ERROR", "WARNING")
                               new_score_policy: str = "only_empty"
                               ) -> None:
    """
    Merge the students and submissions from the source gradebook into the target gradebook

    Assumptions:
    - the target gradebook already contains the assignments
    - new_score_policy: can only be 'only_empty', 'force_new_score', 'only_greater'
    """
    for student in source.students:
        args, kwargs = to_args(student, ['id'])
        target.update_or_create_student(*args, **kwargs)

    for assignment in source.assignments:
        for submission in source.assignment_submissions(assignment.name):
            args, kwargs = to_args(submission, ['student'])
            del kwargs['first_name']
            del kwargs['last_name']
            target.update_or_create_submission(assignment.name, *args, **kwargs)

            for notebook in submission.notebooks:
                for grade in notebook.grades:
                    args, kwargs = to_args(grade, ['name', 'notebook', 'assignment', 'student'])
                    try:
                        target_grade = target.find_grade(*args)
                    except MissingEntry:
                        if on_inconsistency == "WARNING":
                            log = getLogger()
                            log.warning(f"Skipping grade which does not exist in the target: {grade}")
                            continue
                        else:
                            raise

                    if back:
                        grade, target_grade = target_grade, grade
                        args, kwargs = to_args(grade, ['name', 'notebook', 'assignment', 'student'])

                    for key, value in kwargs.items():
                        if ((new_score_policy=="only_empty" or new_score_policy=="only_greater") and getattr(target_grade, key) is None) \
                                or new_score_policy=="force_new_score" \
                                or (new_score_policy=="only_greater" and type(getattr(target_grade, key)) is float and type(getattr(grade, key)) is float and getattr(grade, key) > getattr(target_grade, key)):
                            setattr(target_grade, key, value)

                for comment in notebook.comments:
                    args, kwargs = to_args(comment, ['name', 'notebook', 'assignment', 'student'])
                    try:
                        target_comment = target.find_comment(*args)
                    except MissingEntry:
                        if on_inconsistency == "WARNING" \
                           or (comment.auto_comment is None
                               and comment.manual_comment is None):
                            log = getLogger()
                            log.warning(f"Skipping comment which does not exist in the target: {comment}")
                            continue
                        else:
                            raise

                    if back:
                        comment, target_comment = target_comment, comment
                        args, kwargs = to_args(comment, ['name', 'notebook', 'assignment', 'student'])

                    for key, value in kwargs.items():
                        if getattr(target_comment, key) is None:
                            setattr(target_comment, key, value)

def remove_submission_gradebook(gb: Gradebook,
                                assignment_name: str,
                                student: str) -> None:
    log = getLogger()
    try:
        gb.find_submission(assignment_name, student)
        answer = input(f'Do you really want to suppress {assignment_name} submission of {student} '
                       f'from your local .gradebook.db ? [y/N] ')
        if answer != "y":
            log.info("Submission not deleted from your local .gradebook.db.")
            return
        gb.remove_submission(assignment_name, student)
    except (MissingEntry):
        log.info(f"Missing entry for {assignment_name} submission from student {student} in your local .gradebook.db")


def remove_assignment_gradebook(gb: Gradebook, assignment_name: str, force: bool = False) -> None:
    log = getLogger()
    try:
        gb.find_assignment(assignment_name)
        if not force:
            answer = input(f'Do you really want to suppress all submissions from {assignment_name} '
                           f'from your local .gradebook.db ? [y/N] ')
            if answer != "y":
                log.info("Assignment not deleted from your local .gradebook.db.")
                return

        assignment = gb.find_assignment(assignment_name)

        for submission in assignment.submissions:
            gb.remove_submission(assignment_name, submission.student.id)

        try:
            for notebook in assignment.notebooks:
                gb.remove_notebook(notebook.name, assignment_name)
        except:
            pass

        gb.db.delete(assignment)

        try:
            gb.db.commit()
        except (IntegrityError, FlushError, StatementError) as e:
            app_log.exception("Rolling back session due to database error %s" % e)
            gb.db.rollback()
            raise InvalidEntry(*e.args)
        gb.db.close()

    except (MissingEntry):
        log.info(f"Missing entry for assignment {assignment_name} in your local .gradebook.db")


class GradebookExporter:
    def record(self,
               student: str, assignment: str, notebook_name: str,
               auto_score: float, max_auto_score: float,
               manual_score: Optional[float], max_manual_score: float,
               extra_credit: float) -> None:
        pass

    def record_assignment(self,
                          student: str, assignment: str,
                          auto_score: float, max_auto_score: float,
                          manual_score: Optional[float], max_manual_score: float,
                          extra_credit: float) -> None:
        pass

    def export(self) -> Any:
        pass

class DataFrameGradebookExporter(GradebookExporter):
    def __init__(self) -> None:
        self.data : List[List] = []

    def record(self,
               student: str, assignment: str, notebook_name: str,
               auto_score: float, max_auto_score: float,
               manual_score: Optional[float], max_manual_score: float,
               extra_credit: float) -> None:
        import pandas as pd  # type: ignore
        if auto_score is None:
            auto_score = pd.NA
        if manual_score is None:
            manual_score = pd.NA
        total_score = auto_score
        if max_manual_score > 0:
            total_score += manual_score
        total_score += extra_credit
        self.data.append(
            [student, assignment, notebook_name,
             auto_score, max_auto_score,
             manual_score, max_manual_score,
             extra_credit,
             total_score,
             max_auto_score + max_manual_score,
             ])

    def record_assignment(self,
                          student: str, assignment: str,
                          auto_score: float, max_auto_score: float,
                          manual_score: Optional[float], max_manual_score: float,
                          extra_credit: float) -> None:
        pass

    def export(self) -> Any:
        import pandas as pd   # type: ignore
        df = pd.DataFrame(self.data,
                          columns=['student', 'assignment', 'notebook',
                                   'auto_score', 'max_auto_score',
                                   'manual_score', 'max_manual_score',
                                   'extra_credit',
                                   'total_score', 'max_total_score',
                                   ])
        df.set_index(['student', 'assignment', 'notebook'], inplace=True)
        # df = df.astype('Int64')
        return df


class CSVGradebookExporter(DataFrameGradebookExporter):
    def export(self) -> Any:
        return super().export().to_csv()


class BadgeGradebookExporter(GradebookExporter):

    def record_assignment(self,
                          student: str, assignment: str,
                          auto_score: float, max_auto_score: float,
                          manual_score: Optional[float], max_manual_score: float,
                          extra_credit: float) -> None:

        import anybadge  # type: ignore

        s = f"auto: {auto_score:g}/{max_auto_score:g}"
        if manual_score is None:
            score = auto_score
            max_score = max_auto_score
        else:
            score = auto_score + manual_score
            max_score = max_auto_score + max_manual_score
            s += f" | manual: {manual_score:g}/{max_manual_score:g}"
            s += f" | total: {score:g}/{max_score:g}"

        if extra_credit > 0:
            s += f" | bonus: {extra_credit:g}"

        score += extra_credit
        if max_score > 0:
            score = score / max_score
            if score <= 0.25:
                color = 'red'
            elif score <= 0.5:
                color = 'orange'
            elif score <= 0.75:
                color = 'yellow'
            else:
                color = 'green'
        else:
            color = 'green'

        self.badge = anybadge.Badge("score", s, default_color=color)

    def export(self) -> Any:
        return self.badge


class SVGGradebookExporter(BadgeGradebookExporter):
    def export(self) -> Any:
        return super().export().badge_svg_text


class FormatedGradebookExporter(GradebookExporter):

    header_format: str
    notebook_format: str
    score_format: str
    footer_format: str

    def __init__(self) -> None:
        self.report = self.header_format

    def record(self,
               student: str, assignment: str, notebook_name: str,
               auto_score: float, max_auto_score: float,
               manual_score: Optional[float], max_manual_score: float,
               extra_credit: float) -> None:
        self.report += self.notebook_format.format(
            name=notebook_name)
        if max_auto_score > 0:
            self.report += self.score_format.format(
                type="automatique",
                score=auto_score,
                max_score=max_auto_score)
        if manual_score is not None:
            self.report += self.score_format.format(
                type="manuel",
                score=manual_score,
                max_score=max_manual_score)
        if extra_credit > 0:
            self.report += self.score_format.format(
                type="bonus",
                score=extra_credit,
                max_score=0)

    def record_assignment(self,
                          student: str, assignment: str,
                          auto_score: float, max_auto_score: float,
                          manual_score: Optional[float], max_manual_score: float,
                          extra_credit: float) -> None:
        self.record(
            student, assignment, "total",
            auto_score, max_auto_score,
            manual_score, max_manual_score,
            extra_credit)

    def export(self) -> str:
        return self.report + self.footer_format


class HTMLGradebookExporter(FormatedGradebookExporter):
    header_format = "<html>\n<body>\n<title>Scores</title>\n"
    "<ul>",
    notebook_format = "    <li><a href='{name}.html'>{name}</a>:\n"
    score_format = "        {type}: {score:g}/{max_score:g}\n"
    footer_format = "</ul>\n</body>\n</html>\n"


class MDGradebookExporter(FormatedGradebookExporter):
    header_format = "Scores:\n"
    notebook_format = "- {name}:\n"
    score_format = "     - {type}: {score:g}/{max_score:g}\n"
    footer_format = ""


gradebook_exporters = {
    "md": MDGradebookExporter,
    "html": HTMLGradebookExporter,
    "df": DataFrameGradebookExporter,
    "csv": CSVGradebookExporter,
    "badge": BadgeGradebookExporter,
    "svg": SVGGradebookExporter,
}


def export_scores(gradebook: Gradebook,
                  format: str = "html",
                  student: Optional[str] = None,
                  assignment: Optional[str] = None) -> Any:

    gradebook_exporter: GradebookExporter = gradebook_exporters[format]()

    if student is None:
        students = [student.id for student in gradebook.students]
    else:
        students = [student]

    for student in students:
        assert student is not None
        for submission in gradebook.find_student(student).submissions:
            if assignment is not None and submission.assignment.name != assignment:
                continue
            assignment_auto_score = 0
            assignment_max_auto_score = 0
            assignment_manual_score = None
            assignment_max_manual_score = 0
            assignment_extra_credit = 0
            for notebook in submission.notebooks:
                auto_score = 0
                max_auto_score = 0
                manual_score = None
                max_manual_score = 0
                extra_credit = 0
                for grade in notebook.grades:
                    # Try to guess whether this is a test cell, whose score can be computed automatically
                    # At some point, auto_score was None in this case, but this seems to be gone
                    # For now, assumes that a code cell with a non trivial max_score is a test cell
                    if grade.cell_type == 'code' and grade.auto_score is not None:
                        # Autograded cell
                        max_auto_score += grade.max_score
                        if grade.manual_score is not None:
                            auto_score += grade.manual_score
                        else:
                            auto_score += grade.auto_score
                    else:
                        max_manual_score += grade.max_score
                        if grade.manual_score is not None:
                            if manual_score is None:
                                manual_score = grade.manual_score
                            else:
                                manual_score += grade.manual_score
                    if grade.extra_credit is not None:
                        extra_credit += grade.extra_credit
                gradebook_exporter.record(
                    student, submission.assignment.name,
                    notebook.name,
                    auto_score, max_auto_score,
                    manual_score, max_manual_score,
                    extra_credit)
                assignment_auto_score += auto_score
                assignment_max_auto_score += max_auto_score
                if manual_score is not None:
                    if assignment_manual_score is None:
                        assignment_manual_score = manual_score
                    else:
                        assignment_manual_score += manual_score
                assignment_max_manual_score += max_manual_score
                assignment_extra_credit += extra_credit
            gradebook_exporter.record_assignment(
                    student, submission.assignment.name,
                    assignment_auto_score, assignment_max_auto_score,
                    assignment_manual_score, assignment_max_manual_score,
                    assignment_extra_credit)

    return gradebook_exporter.export()
